import copy
import datetime
import json
import re
import sys
import uuid

from pandajedi.jedicore import Interaction, JediException
from pandajedi.jedicore.JediDatasetSpec import JediDatasetSpec
from pandajedi.jedicore.JediFileSpec import JediFileSpec
from pandajedi.jedicore.JediTaskSpec import JediTaskSpec
from pandaserver.taskbuffer import EventServiceUtils

from . import RefinerUtils

try:
    import idds.common.constants
    import idds.common.utils
    from idds.client.client import Client as iDDS_Client
except ImportError:
    pass


# base class for task refine
class TaskRefinerBase(object):
    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        self.ddmIF = ddmIF
        self.taskBufferIF = taskBufferIF
        self.initializeRefiner(None)
        self.refresh()

    # refresh
    def refresh(self):
        self.siteMapper = self.taskBufferIF.getSiteMapper()

    # initialize
    def initializeRefiner(self, tmpLog):
        self.taskSpec = None
        self.inMasterDatasetSpec = []
        self.inSecDatasetSpecList = []
        self.outDatasetSpecList = []
        self.outputTemplateMap = {}
        self.jobParamsTemplate = None
        self.cloudName = None
        self.siteName = None
        self.tmpLog = tmpLog
        self.updatedTaskParams = None
        self.unmergeMasterDatasetSpec = {}
        self.unmergeDatasetSpecMap = {}
        self.oldTaskStatus = None
        self.unknownDatasetList = []

    # set jobParamsTemplate
    def setJobParamsTemplate(self, jobParamsTemplate):
        self.jobParamsTemplate = jobParamsTemplate

    # extract common parameters
    def extractCommon(self, jediTaskID, taskParamMap, workQueueMapper, splitRule):
        # remove irrelevant
        if "maxAttempt" in taskParamMap and not taskParamMap["maxAttempt"]:
            del taskParamMap["maxAttempt"]
        # make task spec
        taskSpec = JediTaskSpec()
        taskSpec.jediTaskID = jediTaskID
        taskSpec.attemptNr = 0
        taskSpec.taskName = taskParamMap["taskName"]
        taskSpec.userName = taskParamMap["userName"]
        taskSpec.vo = taskParamMap["vo"]
        taskSpec.prodSourceLabel = taskParamMap["prodSourceLabel"]
        taskSpec.taskPriority = taskParamMap["taskPriority"]
        if taskSpec.taskPriority is None:
            taskSpec.taskPriority = 0
        if "currentPriority" in taskParamMap:
            taskSpec.currentPriority = taskParamMap["currentPriority"]
        else:
            taskSpec.currentPriority = taskSpec.taskPriority
        taskSpec.architecture = taskParamMap["architecture"]
        taskSpec.transUses = taskParamMap["transUses"]
        taskSpec.transHome = taskParamMap["transHome"]
        if "transPath" in taskParamMap:
            taskSpec.transPath = taskParamMap["transPath"]
        taskSpec.processingType = taskParamMap["processingType"]
        taskSpec.taskType = taskParamMap["taskType"]
        taskSpec.splitRule = splitRule
        taskSpec.startTime = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
        if "workingGroup" in taskParamMap:
            taskSpec.workingGroup = taskParamMap["workingGroup"]
        if "countryGroup" in taskParamMap:
            taskSpec.countryGroup = taskParamMap["countryGroup"]
        if "ticketID" in taskParamMap:
            taskSpec.ticketID = taskParamMap["ticketID"]
        if "ticketSystemType" in taskParamMap:
            taskSpec.ticketSystemType = taskParamMap["ticketSystemType"]
        if "reqID" in taskParamMap:
            taskSpec.reqID = taskParamMap["reqID"]
        else:
            taskSpec.reqID = jediTaskID
        if "coreCount" in taskParamMap:
            taskSpec.coreCount = taskParamMap["coreCount"]
        else:
            taskSpec.coreCount = 1
        if "walltime" in taskParamMap:
            taskSpec.walltime = taskParamMap["walltime"]
        else:
            taskSpec.walltime = 0
        if "walltimeUnit" not in taskParamMap:
            # force to set NULL so that retried tasks get data from scouts again
            taskSpec.forceUpdate("walltimeUnit")
        if "outDiskCount" in taskParamMap:
            taskSpec.outDiskCount = taskParamMap["outDiskCount"]
        else:
            taskSpec.outDiskCount = 0
        if "outDiskUnit" in taskParamMap:
            taskSpec.outDiskUnit = taskParamMap["outDiskUnit"]
        if "workDiskCount" in taskParamMap:
            taskSpec.workDiskCount = taskParamMap["workDiskCount"]
        else:
            taskSpec.workDiskCount = 0
        if "workDiskUnit" in taskParamMap:
            taskSpec.workDiskUnit = taskParamMap["workDiskUnit"]
        if "ramCount" in taskParamMap:
            taskSpec.ramCount = taskParamMap["ramCount"]
        else:
            taskSpec.ramCount = 0
        if "ramUnit" in taskParamMap:
            taskSpec.ramUnit = taskParamMap["ramUnit"]
        elif "ramCountUnit" in taskParamMap:
            taskSpec.ramUnit = taskParamMap["ramCountUnit"]
        if "baseRamCount" in taskParamMap:
            taskSpec.baseRamCount = taskParamMap["baseRamCount"]
        else:
            taskSpec.baseRamCount = 0
        # IO
        if "ioIntensity" in taskParamMap:
            taskSpec.ioIntensity = taskParamMap["ioIntensity"]
        if "ioIntensityUnit" in taskParamMap:
            taskSpec.ioIntensityUnit = taskParamMap["ioIntensityUnit"]
        # HS06 stuff
        if "cpuTimeUnit" in taskParamMap:
            taskSpec.cpuTimeUnit = taskParamMap["cpuTimeUnit"]
        if "cpuTime" in taskParamMap:
            taskSpec.cpuTime = taskParamMap["cpuTime"]
        if "cpuEfficiency" in taskParamMap:
            taskSpec.cpuEfficiency = taskParamMap["cpuEfficiency"]
        else:
            # 90% of cpu efficiency by default
            taskSpec.cpuEfficiency = 90
        if "baseWalltime" in taskParamMap:
            taskSpec.baseWalltime = taskParamMap["baseWalltime"]
        else:
            # 10min of offset by default
            taskSpec.baseWalltime = 10 * 60
        # for merge
        if "mergeRamCount" in taskParamMap:
            taskSpec.mergeRamCount = taskParamMap["mergeRamCount"]
        if "mergeCoreCount" in taskParamMap:
            taskSpec.mergeCoreCount = taskParamMap["mergeCoreCount"]
        # scout
        if "skipScout" not in taskParamMap and not taskSpec.isPostScout():
            taskSpec.setUseScout(True)
        # cloud
        if "cloud" in taskParamMap:
            self.cloudName = taskParamMap["cloud"]
            taskSpec.cloud = self.cloudName
        else:
            # set dummy to force update
            taskSpec.cloud = "dummy"
            taskSpec.cloud = None
        # site
        if "site" in taskParamMap:
            self.siteName = taskParamMap["site"]
            taskSpec.site = self.siteName
        else:
            # set dummy to force update
            taskSpec.site = "dummy"
            taskSpec.site = None
        # nucleus
        if "nucleus" in taskParamMap:
            taskSpec.nucleus = taskParamMap["nucleus"]
        # preset some parameters for job cloning
        if "useJobCloning" in taskParamMap:
            # set implicit parameters
            if "nEventsPerWorker" not in taskParamMap:
                taskParamMap["nEventsPerWorker"] = 1
            if "nSitesPerJob" not in taskParamMap:
                taskParamMap["nSitesPerJob"] = 2
            if "nEsConsumers" not in taskParamMap:
                taskParamMap["nEsConsumers"] = taskParamMap["nSitesPerJob"]
        # minimum granularity
        if "minGranularity" in taskParamMap:
            taskParamMap["nEventsPerRange"] = taskParamMap["minGranularity"]
        # event service flag
        if "useJobCloning" in taskParamMap:
            taskSpec.eventService = 2
        elif "nEventsPerWorker" in taskParamMap:
            taskSpec.eventService = 1
        elif "fineGrainedProc" in taskParamMap:
            taskSpec.eventService = 3
        else:
            taskSpec.eventService = 0
        # OS
        if "osInfo" in taskParamMap:
            taskSpec.termCondition = taskParamMap["osInfo"]
        # ttcr: requested time to completion
        if "ttcrTimestamp" in taskParamMap:
            try:
                # get rid of the +00:00 timezone string and parse the timestamp
                taskSpec.ttcRequested = datetime.datetime.strptime(taskParamMap["ttcrTimestamp"].split("+")[0], "%Y-%m-%d %H:%M:%S.%f")
            except (IndexError, ValueError):
                pass
        # goal
        if "goal" in taskParamMap:
            try:
                taskSpec.goal = int(float(taskParamMap["goal"]) * 10)
                if taskSpec.goal > 1000:
                    taskSpec.goal = None
            except Exception:
                pass
        # campaign
        if "campaign" in taskParamMap:
            taskSpec.campaign = taskParamMap["campaign"]
        # request type
        if "requestType" in taskParamMap:
            taskSpec.requestType = taskParamMap["requestType"]
        # image name
        if "container_name" in taskParamMap:
            taskSpec.container_name = taskParamMap["container_name"]
        self.taskSpec = taskSpec
        # set split rule
        if "nFilesPerJob" not in taskParamMap:
            if "tgtNumEventsPerJob" in taskParamMap:
                # set nEventsPerJob not to respect file boundaries when nFilesPerJob is not used
                self.setSplitRule(None, taskParamMap["tgtNumEventsPerJob"], JediTaskSpec.splitRuleToken["nEventsPerJob"])
            elif (
                "nEventsPerInputFile" in taskParamMap
                and "nEventsPerJob" in taskParamMap
                and taskParamMap["nEventsPerJob"] >= taskParamMap["nEventsPerInputFile"]
            ):
                # set nFilesPerJob if nEventsPerJob and nEventsPerInputFile are set
                nFilesPerJob = taskParamMap["nEventsPerJob"] // taskParamMap["nEventsPerInputFile"]
                self.setSplitRule(None, nFilesPerJob, JediTaskSpec.splitRuleToken["nFilesPerJob"])
        self.setSplitRule(taskParamMap, "nFilesPerJob", JediTaskSpec.splitRuleToken["nFilesPerJob"])
        self.setSplitRule(taskParamMap, "nEventsPerJob", JediTaskSpec.splitRuleToken["nEventsPerJob"])
        self.setSplitRule(taskParamMap, "nGBPerJob", JediTaskSpec.splitRuleToken["nGBPerJob"])
        self.setSplitRule(taskParamMap, "nMaxFilesPerJob", JediTaskSpec.splitRuleToken["nMaxFilesPerJob"])
        self.setSplitRule(taskParamMap, "maxEventsPerJob", JediTaskSpec.splitRuleToken["maxEventsPerJob"])
        self.setSplitRule(taskParamMap, "nEventsPerWorker", JediTaskSpec.splitRuleToken["nEventsPerWorker"])
        self.setSplitRule(taskParamMap, "nEventsPerInputFile", JediTaskSpec.splitRuleToken["nEventsPerInput"])
        self.setSplitRule(taskParamMap, "disableAutoRetry", JediTaskSpec.splitRuleToken["disableAutoRetry"])
        self.setSplitRule(taskParamMap, "nEsConsumers", JediTaskSpec.splitRuleToken["nEsConsumers"])
        self.setSplitRule(taskParamMap, "waitInput", JediTaskSpec.splitRuleToken["waitInput"])
        self.setSplitRule(taskParamMap, "addNthFieldToLFN", JediTaskSpec.splitRuleToken["addNthFieldToLFN"])
        self.setSplitRule(taskParamMap, "scoutSuccessRate", JediTaskSpec.splitRuleToken["scoutSuccessRate"])
        self.setSplitRule(taskParamMap, "t1Weight", JediTaskSpec.splitRuleToken["t1Weight"])
        self.setSplitRule(taskParamMap, "maxAttemptES", JediTaskSpec.splitRuleToken["maxAttemptES"])
        self.setSplitRule(taskParamMap, "maxAttemptEsJob", JediTaskSpec.splitRuleToken["maxAttemptEsJob"])
        self.setSplitRule(taskParamMap, "nSitesPerJob", JediTaskSpec.splitRuleToken["nSitesPerJob"])
        self.setSplitRule(taskParamMap, "nEventsPerMergeJob", JediTaskSpec.splitRuleToken["nEventsPerMergeJob"])
        self.setSplitRule(taskParamMap, "nFilesPerMergeJob", JediTaskSpec.splitRuleToken["nFilesPerMergeJob"])
        self.setSplitRule(taskParamMap, "nGBPerMergeJob", JediTaskSpec.splitRuleToken["nGBPerMergeJob"])
        self.setSplitRule(taskParamMap, "nMaxFilesPerMergeJob", JediTaskSpec.splitRuleToken["nMaxFilesPerMergeJob"])
        self.setSplitRule(taskParamMap, "maxWalltime", JediTaskSpec.splitRuleToken["maxWalltime"])
        self.setSplitRule(taskParamMap, "tgtMaxOutputForNG", JediTaskSpec.splitRuleToken["tgtMaxOutputForNG"])
        self.setSplitRule(taskParamMap, "maxNumJobs", JediTaskSpec.splitRuleToken["maxNumJobs"])
        self.setSplitRule(taskParamMap, "totNumJobs", JediTaskSpec.splitRuleToken["totNumJobs"])
        self.setSplitRule(taskParamMap, "nChunksToWait", JediTaskSpec.splitRuleToken["nChunksToWait"])
        self.setSplitRule(taskParamMap, "retryRamOffset", JediTaskSpec.splitRuleToken["retryRamOffset"])
        self.setSplitRule(taskParamMap, "retryRamStep", JediTaskSpec.splitRuleToken["retryRamStep"])
        if "forceStaged" in taskParamMap:
            taskParamMap["useLocalIO"] = taskParamMap["forceStaged"]
        if "useLocalIO" in taskParamMap:
            if taskParamMap["useLocalIO"]:
                self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["useLocalIO"])
            else:
                self.setSplitRule(None, 0, JediTaskSpec.splitRuleToken["useLocalIO"])
        if "nJumboJobs" in taskParamMap:
            self.setSplitRule(taskParamMap, "nJumboJobs", JediTaskSpec.splitRuleToken["nJumboJobs"])
            taskSpec.useJumbo = JediTaskSpec.enum_useJumbo["waiting"]
            if "maxJumboPerSite" in taskParamMap:
                self.setSplitRule(taskParamMap, "maxJumboPerSite", JediTaskSpec.splitRuleToken["maxJumboPerSite"])
        if "minCpuEfficiency" in taskParamMap:
            self.setSplitRule(taskParamMap, "minCpuEfficiency", JediTaskSpec.splitRuleToken["minCpuEfficiency"])
        if "loadXML" in taskParamMap:
            self.setSplitRule(None, 3, JediTaskSpec.splitRuleToken["loadXML"])
            self.setSplitRule(None, 4, JediTaskSpec.splitRuleToken["groupBoundaryID"])
        if "pfnList" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["pfnList"])
        if "noWaitParent" in taskParamMap and taskParamMap["noWaitParent"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["noWaitParent"])
        if "respectLB" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["respectLB"])
        if "releasePerLB" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["releasePerLB"])
        if "orderByLB" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["orderByLB"])
        if "respectSplitRule" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["respectSplitRule"])
        if "reuseSecOnDemand" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["reuseSecOnDemand"])
        if "ddmBackEnd" in taskParamMap:
            self.taskSpec.setDdmBackEnd(taskParamMap["ddmBackEnd"])
        if "disableReassign" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["disableReassign"])
        if "allowPartialFinish" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["allowPartialFinish"])
        if "useExhausted" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["useExhausted"])
        if "useRealNumEvents" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["useRealNumEvents"])
        if "ipConnectivity" in taskParamMap:
            self.taskSpec.setIpConnectivity(taskParamMap["ipConnectivity"])
        if "altStageOut" in taskParamMap:
            self.taskSpec.setAltStageOut(taskParamMap["altStageOut"])
        if "allowInputLAN" in taskParamMap:
            self.taskSpec.setAllowInputLAN(taskParamMap["allowInputLAN"])
        if "runUntilClosed" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["runUntilClosed"])
        if "stayOutputOnSite" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["stayOutputOnSite"])
        if "useJobCloning" in taskParamMap:
            scValue = EventServiceUtils.getJobCloningValue(taskParamMap["useJobCloning"])
            self.setSplitRule(None, scValue, JediTaskSpec.splitRuleToken["useJobCloning"])
        if "failWhenGoalUnreached" in taskParamMap and taskParamMap["failWhenGoalUnreached"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["failGoalUnreached"])
        if "switchEStoNormal" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["switchEStoNormal"])
        if "nEventsPerRange" in taskParamMap:
            self.setSplitRule(taskParamMap, "nEventsPerRange", JediTaskSpec.splitRuleToken["dynamicNumEvents"])
        if "allowInputWAN" in taskParamMap and taskParamMap["allowInputWAN"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["allowInputWAN"])
        if "putLogToOS" in taskParamMap and taskParamMap["putLogToOS"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["putLogToOS"])
        if "mergeEsOnOS" in taskParamMap and taskParamMap["mergeEsOnOS"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["mergeEsOnOS"])
        if "writeInputToFile" in taskParamMap and taskParamMap["writeInputToFile"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["writeInputToFile"])
        if "useFileAsSourceLFN" in taskParamMap and taskParamMap["useFileAsSourceLFN"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["useFileAsSourceLFN"])
        if "ignoreMissingInDS" in taskParamMap and taskParamMap["ignoreMissingInDS"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["ignoreMissingInDS"])
        if "noExecStrCnv" in taskParamMap and taskParamMap["noExecStrCnv"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["noExecStrCnv"])
        if "inFilePosEvtNum" in taskParamMap and taskParamMap["inFilePosEvtNum"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["inFilePosEvtNum"])
        if self.taskSpec.useEventService() and not taskSpec.useJobCloning():
            if "registerEsFiles" in taskParamMap and taskParamMap["registerEsFiles"] is True:
                self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["registerEsFiles"])
        if "disableAutoFinish" in taskParamMap and taskParamMap["disableAutoFinish"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["disableAutoFinish"])
        if "resurrectConsumers" in taskParamMap and taskParamMap["resurrectConsumers"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["resurrectConsumers"])
        if "usePrefetcher" in taskParamMap and taskParamMap["usePrefetcher"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["usePrefetcher"])
        if "notDiscardEvents" in taskParamMap and taskParamMap["notDiscardEvents"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["notDiscardEvents"])
        if "decAttOnFailedES" in taskParamMap and taskParamMap["decAttOnFailedES"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["decAttOnFailedES"])
        if "useZipToPin" in taskParamMap and taskParamMap["useZipToPin"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["useZipToPin"])
        if "osMatching" in taskParamMap and taskParamMap["osMatching"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["osMatching"])
        if "multiStepExec" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["multiStepExec"])
        if "onlyTagsForFC" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["onlyTagsForFC"])
        if "segmentedWork" in taskParamMap and "segmentSpecs" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["segmentedWork"])
        if "avoidVP" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["avoidVP"])
        if "inputPreStaging" in taskParamMap and taskParamMap["inputPreStaging"] is True:
            self.setSplitRule(None, JediTaskSpec.enum_inputPreStaging["use"], JediTaskSpec.splitRuleToken["inputPreStaging"])
        if "hpoWorkflow" in taskParamMap and taskParamMap["hpoWorkflow"] is True and "hpoRequestData" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["hpoWorkflow"])
        if "noLoopingCheck" in taskParamMap and taskParamMap["noLoopingCheck"]:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["noLoopingCheck"])
        if "encJobParams" in taskParamMap and taskParamMap["encJobParams"]:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["encJobParams"])
        if "useSecrets" in taskParamMap and taskParamMap["useSecrets"]:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["useSecrets"])
        if "debugMode" in taskParamMap and taskParamMap["debugMode"]:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["debugMode"])
        if "pushStatusChanges" in taskParamMap and taskParamMap["pushStatusChanges"]:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["pushStatusChanges"])
        if "maxCoreCount" in taskParamMap:
            self.setSplitRule(taskParamMap, "maxCoreCount", JediTaskSpec.splitRuleToken["maxCoreCount"])
        if "cloudAsVO" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["cloudAsVO"])
        if "pushJob" in taskParamMap and taskParamMap["pushJob"]:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["pushJob"])
        if "fineGrainedProc" in taskParamMap and taskParamMap["fineGrainedProc"]:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["fineGrainedProc"])
        if "onSiteMerging" in taskParamMap and taskParamMap["onSiteMerging"]:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["onSiteMerging"])
        if "fullChain" in taskParamMap:
            self.taskSpec.set_full_chain(taskParamMap["fullChain"])
        if "orderInputBy" in taskParamMap:
            self.taskSpec.set_order_input_by(taskParamMap["orderInputBy"])
        if "intermediateTask" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["intermediateTask"])
        if "allowEmptyInput" in taskParamMap:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["allowEmptyInput"])
        if "messageDriven" in taskParamMap and taskParamMap["messageDriven"]:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["messageDriven"])
        # work queue
        workQueue = None
        if "workQueueName" in taskParamMap:
            # work queue is specified
            workQueue = workQueueMapper.getQueueByName(taskSpec.vo, taskSpec.prodSourceLabel, taskParamMap["workQueueName"])
        if workQueue is None:
            # get work queue based on task attributes
            workQueue, tmpStr = workQueueMapper.getQueueWithSelParams(
                taskSpec.vo,
                taskSpec.prodSourceLabel,
                prodSourceLabel=taskSpec.prodSourceLabel,
                processingType=taskSpec.processingType,
                workingGroup=taskSpec.workingGroup,
                coreCount=taskSpec.coreCount,
                site=taskSpec.site,
                eventService=taskSpec.eventService,
                splitRule=taskSpec.splitRule,
                campaign=taskSpec.campaign,
            )
        if workQueue is None:
            errStr = f"workqueue is undefined for vo={taskSpec.vo} label={taskSpec.prodSourceLabel} "
            errStr += "processingType={0} workingGroup={1} coreCount={2} eventService={3} ".format(
                taskSpec.processingType, taskSpec.workingGroup, taskSpec.coreCount, taskSpec.eventService
            )
            errStr += f"splitRule={taskSpec.splitRule} campaign={taskSpec.campaign}"
            raise RuntimeError(errStr)
        self.taskSpec.workQueue_ID = workQueue.queue_id

        # Initialize the global share
        gshare = "Undefined"
        if "gshare" in taskParamMap and self.taskBufferIF.is_valid_share(taskParamMap["gshare"]):
            # work queue is specified
            gshare = taskParamMap["gshare"]
        else:
            # get share based on definition
            gshare = self.taskBufferIF.get_share_for_task(self.taskSpec)
            if gshare is None:
                gshare = "Undefined"  # Should not happen. Undefined is set when no share is found
                # errStr  = 'share is undefined for vo={0} label={1} '.format(taskSpec.vo,taskSpec.prodSourceLabel)
                # errStr += 'workingGroup={0} campaign={1} '.format(taskSpec.workingGroup, taskSpec.campaign)
                # raise RuntimeError,errStr
        self.taskSpec.gshare = gshare

        # Initialize the resource type
        try:
            self.taskSpec.resource_type = self.taskBufferIF.get_resource_type_task(self.taskSpec)
        except Exception:
            self.taskSpec.resource_type = "Undefined"

        # return
        return

    # basic refinement procedure
    def doBasicRefine(self, taskParamMap):
        # get input/output/log dataset specs
        nIn = 0
        nOutMap = {}
        if "log" not in taskParamMap:
            itemList = taskParamMap["jobParameters"]
        elif isinstance(taskParamMap["log"], dict):
            itemList = taskParamMap["jobParameters"] + [taskParamMap["log"]]
        else:
            itemList = taskParamMap["jobParameters"] + taskParamMap["log"]
        if "log_merge" in taskParamMap:
            itemList += [taskParamMap["log_merge"]]
        # pseudo input
        if "noInput" in taskParamMap and taskParamMap["noInput"] is True:
            tmpItem = {}
            tmpItem["type"] = "template"
            tmpItem["value"] = ""
            tmpItem["dataset"] = "pseudo_dataset"
            tmpItem["param_type"] = "pseudo_input"
            itemList = [tmpItem] + itemList
        # random seed
        if RefinerUtils.useRandomSeed(taskParamMap):
            tmpItem = {}
            tmpItem["type"] = "template"
            tmpItem["value"] = ""
            tmpItem["dataset"] = "RNDMSEED"
            tmpItem["param_type"] = "random_seed"
            itemList.append(tmpItem)
        # loop over all items
        allDsList = []
        for tmpItem in itemList:
            # look for datasets
            if tmpItem["type"] == "template" and "dataset" in tmpItem:
                # avoid duplication
                if tmpItem["dataset"] not in allDsList:
                    allDsList.append(tmpItem["dataset"])
                else:
                    continue
                datasetSpec = JediDatasetSpec()
                datasetSpec.datasetName = tmpItem["dataset"]
                datasetSpec.jediTaskID = self.taskSpec.jediTaskID
                datasetSpec.type = tmpItem["param_type"]
                if "container" in tmpItem:
                    datasetSpec.containerName = tmpItem["container"]
                if "token" in tmpItem:
                    datasetSpec.storageToken = tmpItem["token"]
                if "destination" in tmpItem:
                    datasetSpec.destination = tmpItem["destination"]
                if "attributes" in tmpItem:
                    datasetSpec.setDatasetAttribute(tmpItem["attributes"])
                if "ratio" in tmpItem:
                    datasetSpec.setDatasetAttribute(f"ratio={tmpItem['ratio']}")
                if "eventRatio" in tmpItem:
                    datasetSpec.setEventRatio(tmpItem["eventRatio"])
                if "check" in tmpItem:
                    datasetSpec.setDatasetAttribute("cc")
                if "usedup" in tmpItem:
                    datasetSpec.setDatasetAttribute("ud")
                if "random" in tmpItem:
                    datasetSpec.setDatasetAttribute("rd")
                if "reusable" in tmpItem:
                    datasetSpec.setDatasetAttribute("ru")
                if "indexConsistent" in tmpItem:
                    datasetSpec.setDatasetAttributeWithLabel("indexConsistent")
                if "mergeOnly" in tmpItem:
                    datasetSpec.setDatasetAttributeWithLabel("mergeOnly")
                if "offset" in tmpItem:
                    datasetSpec.setOffset(tmpItem["offset"])
                if "allowNoOutput" in tmpItem:
                    datasetSpec.allowNoOutput()
                if "nFilesPerJob" in tmpItem:
                    datasetSpec.setNumFilesPerJob(tmpItem["nFilesPerJob"])
                if "num_records" in tmpItem:
                    datasetSpec.setNumRecords(tmpItem["num_records"])
                if "transient" in tmpItem:
                    datasetSpec.setTransient(tmpItem["transient"])
                if "pseudo" in tmpItem:
                    datasetSpec.setPseudo()
                datasetSpec.vo = self.taskSpec.vo
                datasetSpec.nFiles = 0
                datasetSpec.nFilesUsed = 0
                datasetSpec.nFilesFinished = 0
                datasetSpec.nFilesFailed = 0
                datasetSpec.nFilesOnHold = 0
                datasetSpec.nFilesWaiting = 0
                datasetSpec.nFilesMissing = 0
                datasetSpec.nEvents = 0
                datasetSpec.nEventsUsed = 0
                datasetSpec.nEventsToBeUsed = 0
                datasetSpec.status = "defined"
                if datasetSpec.type in JediDatasetSpec.getInputTypes() + ["random_seed"]:
                    datasetSpec.streamName = RefinerUtils.extractStreamName(tmpItem["value"])
                    if "expandedList" not in tmpItem:
                        tmpItem["expandedList"] = []
                    # dataset names could be comma-concatenated
                    datasetNameList = datasetSpec.datasetName.split(",")
                    # datasets could be added by incexec
                    incexecDS = f"dsFor{datasetSpec.streamName}"
                    # remove /XYZ
                    incexecDS = incexecDS.split("/")[0]
                    if incexecDS in taskParamMap:
                        for tmpDatasetName in taskParamMap[incexecDS].split(","):
                            if tmpDatasetName not in datasetNameList:
                                datasetNameList.append(tmpDatasetName)
                    # consolidation
                    if len(datasetNameList) > 1 and "consolidate" in tmpItem:
                        tmpIF = self.ddmIF.getInterface(self.taskSpec.vo, self.taskSpec.cloud)
                        if tmpIF:
                            containerName = tmpItem["consolidate"]
                            tmpStat = tmpIF.registerNewDataset(containerName)
                            if not tmpStat:
                                errStr = f"failed to register {containerName}"
                                raise JediException.ExternalTempError(errStr)
                            tmpDsListInCont = tmpIF.listDatasetsInContainer(containerName)
                            for tmpContName in datasetNameList:
                                tmpDsNameList = tmpIF.expandContainer(tmpContName)
                                for tmpDsName in tmpDsNameList:
                                    if tmpDsName not in tmpDsListInCont:
                                        tmpStat = tmpIF.addDatasetsToContainer(containerName, [tmpDsName])
                                        if not tmpStat:
                                            errStr = f"failed to add {tmpDsName} to {containerName}"
                                            raise JediException.ExternalTempError(errStr)
                            datasetNameList = [containerName]
                    # loop over all dataset names
                    inDatasetSpecList = []
                    for datasetName in datasetNameList:
                        # skip empty
                        if datasetName == "":
                            continue
                        # expand
                        if datasetSpec.isPseudo() or datasetSpec.type in ["random_seed"] or datasetName == "DBR_LATEST":
                            # pseudo input
                            tmpDatasetNameList = [datasetName]
                            if self.taskSpec.is_work_segmented():
                                tmpDatasetNameList *= len(taskParamMap["segmentSpecs"])
                        else:
                            tmpIF = self.ddmIF.getInterface(self.taskSpec.vo, self.taskSpec.cloud)
                            if not tmpIF:
                                tmpDatasetNameList = []
                            else:
                                if "expand" in tmpItem and tmpItem["expand"] is True:
                                    # expand dataset container
                                    tmpDatasetNameList = tmpIF.expandContainer(datasetName)
                                else:
                                    # normal dataset name
                                    tmpDatasetNameList = tmpIF.listDatasets(datasetName)
                        i_element = 0
                        for elementDatasetName in tmpDatasetNameList:
                            if nIn > 0 or elementDatasetName not in tmpItem["expandedList"] or self.taskSpec.is_work_segmented():
                                tmpItem["expandedList"].append(elementDatasetName)
                                inDatasetSpec = copy.copy(datasetSpec)
                                inDatasetSpec.datasetName = elementDatasetName
                                if nIn > 0 or not self.taskSpec.is_hpo_workflow():
                                    inDatasetSpec.containerName = datasetName
                                else:
                                    if self.taskSpec.is_work_segmented():
                                        inDatasetSpec.containerName = "{}/{}".format(
                                            taskParamMap["segmentSpecs"][i_element]["name"], taskParamMap["segmentSpecs"][i_element]["id"]
                                        )
                                    else:
                                        inDatasetSpec.containerName = "None/None"
                                inDatasetSpecList.append(inDatasetSpec)
                            i_element += 1
                    # empty input
                    if inDatasetSpecList == [] and self.oldTaskStatus != "rerefine":
                        errStr = f'doBasicRefine : unknown input dataset "{datasetSpec.datasetName}"'
                        self.taskSpec.setErrDiag(errStr)
                        if datasetSpec.datasetName not in self.unknownDatasetList:
                            self.unknownDatasetList.append(datasetSpec.datasetName)
                        raise JediException.UnknownDatasetError(errStr)
                    # set master flag
                    for inDatasetSpec in inDatasetSpecList:
                        if nIn == 0:
                            # master
                            self.inMasterDatasetSpec.append(inDatasetSpec)
                        else:
                            # secondary
                            self.inSecDatasetSpecList.append(inDatasetSpec)
                    nIn += 1
                    continue
                if datasetSpec.type in ["output", "log"]:
                    if datasetSpec.type not in nOutMap:
                        nOutMap[datasetSpec.type] = 0
                    # make stream name
                    if not datasetSpec.is_merge_only():
                        datasetSpec.streamName = f"{datasetSpec.type.upper()}{nOutMap[datasetSpec.type]}"
                    else:
                        datasetSpec.streamName = "LOG_MERGE"
                    nOutMap[datasetSpec.type] += 1
                    # set attribute for event service
                    if self.taskSpec.useEventService() and "objectStore" in taskParamMap and datasetSpec.type in ["output"]:
                        datasetSpec.setObjectStore(taskParamMap["objectStore"])
                    # extract output filename template and change the value field
                    outFileTemplate, tmpItem["value"] = RefinerUtils.extractReplaceOutFileTemplate(tmpItem["value"], datasetSpec.streamName)
                    # make output template
                    if outFileTemplate is not None:
                        if "offset" in tmpItem:
                            offsetVal = 1 + tmpItem["offset"]
                        else:
                            offsetVal = 1
                        outTemplateMap = {
                            "jediTaskID": self.taskSpec.jediTaskID,
                            "serialNr": offsetVal,
                            "streamName": datasetSpec.streamName,
                            "filenameTemplate": outFileTemplate,
                            "outtype": datasetSpec.type,
                        }
                        if datasetSpec.outputMapKey() in self.outputTemplateMap:
                            # multiple files are associated to the same output datasets
                            self.outputTemplateMap[datasetSpec.outputMapKey()].append(outTemplateMap)
                            # don't insert the same output dataset
                            continue
                        self.outputTemplateMap[datasetSpec.outputMapKey()] = [outTemplateMap]
                    # append
                    self.outDatasetSpecList.append(datasetSpec)
                    # used only in merge
                    if datasetSpec.is_merge_only():
                        continue
                    # make unmerged dataset
                    if "mergeOutput" in taskParamMap and taskParamMap["mergeOutput"] is True:
                        umDatasetSpec = JediDatasetSpec()
                        umDatasetSpec.datasetName = "panda.um." + datasetSpec.datasetName
                        umDatasetSpec.jediTaskID = self.taskSpec.jediTaskID
                        umDatasetSpec.storageToken = "TOMERGE"
                        umDatasetSpec.vo = datasetSpec.vo
                        umDatasetSpec.type = "tmpl_trn_" + datasetSpec.type
                        umDatasetSpec.nFiles = 0
                        umDatasetSpec.nFilesUsed = 0
                        umDatasetSpec.nFilesToBeUsed = 0
                        umDatasetSpec.nFilesFinished = 0
                        umDatasetSpec.nFilesFailed = 0
                        umDatasetSpec.nFilesOnHold = 0
                        umDatasetSpec.status = "defined"
                        umDatasetSpec.streamName = datasetSpec.streamName
                        if datasetSpec.isAllowedNoOutput():
                            umDatasetSpec.allowNoOutput()
                        # ratio
                        if datasetSpec.getRatioToMaster() > 1:
                            umDatasetSpec.setDatasetAttribute(f"ratio={datasetSpec.getRatioToMaster()}")
                        # make unmerged output template
                        if outFileTemplate is not None:
                            umOutTemplateMap = {
                                "jediTaskID": self.taskSpec.jediTaskID,
                                "serialNr": 1,
                                "streamName": umDatasetSpec.streamName,
                                "outtype": datasetSpec.type,
                            }
                            # append temporary name
                            if "umNameAtEnd" in taskParamMap and taskParamMap["umNameAtEnd"] is True:
                                # append temporary name at the end
                                umOutTemplateMap["filenameTemplate"] = outFileTemplate + ".panda.um"
                            else:
                                umOutTemplateMap["filenameTemplate"] = "panda.um." + outFileTemplate
                            if umDatasetSpec.outputMapKey() in self.outputTemplateMap:
                                # multiple files are associated to the same output datasets
                                self.outputTemplateMap[umDatasetSpec.outputMapKey()].append(umOutTemplateMap)
                                # don't insert the same output dataset
                                continue
                            self.outputTemplateMap[umDatasetSpec.outputMapKey()] = [umOutTemplateMap]
                        # use log as master for merging
                        if datasetSpec.type == "log":
                            self.unmergeMasterDatasetSpec[datasetSpec.outputMapKey()] = umDatasetSpec
                        else:
                            # append
                            self.unmergeDatasetSpecMap[datasetSpec.outputMapKey()] = umDatasetSpec
        # set attributes for merging
        if "mergeOutput" in taskParamMap and taskParamMap["mergeOutput"] is True:
            self.setSplitRule(None, 1, JediTaskSpec.splitRuleToken["mergeOutput"])
        # make job parameters
        rndmSeedOffset = None
        firstEventOffset = None
        jobParameters = ""
        for tmpItem in taskParamMap["jobParameters"]:
            if "value" in tmpItem:
                # hidden parameter
                if "hidden" in tmpItem and tmpItem["hidden"] is True:
                    continue
                # add tags for ES-only parameters
                esOnly = False
                if "es_only" in tmpItem and tmpItem["es_only"] is True:
                    esOnly = True
                if esOnly:
                    jobParameters += "<PANDA_ES_ONLY>"
                jobParameters += f"{tmpItem['value']}"
                if esOnly:
                    jobParameters += "</PANDA_ES_ONLY>"
                # padding
                if "padding" in tmpItem and tmpItem["padding"] is False:
                    pass
                else:
                    jobParameters += " "
                # get offset for random seed and first event
                if tmpItem["type"] == "template" and tmpItem["param_type"] == "number":
                    if "${RNDMSEED}" in tmpItem["value"]:
                        if "offset" in tmpItem:
                            rndmSeedOffset = tmpItem["offset"]
                        else:
                            rndmSeedOffset = 0
                    elif "${FIRSTEVENT}" in tmpItem["value"]:
                        if "offset" in tmpItem:
                            firstEventOffset = tmpItem["offset"]
        jobParameters = jobParameters[:-1]
        # append parameters for event service merging if necessary
        esmergeParams = self.getParamsForEventServiceMerging(taskParamMap)
        if esmergeParams is not None:
            jobParameters += esmergeParams
        self.setJobParamsTemplate(jobParameters)
        # set random seed offset
        if rndmSeedOffset is not None:
            self.setSplitRule(None, rndmSeedOffset, JediTaskSpec.splitRuleToken["randomSeed"])
        if firstEventOffset is not None:
            self.setSplitRule(None, firstEventOffset, JediTaskSpec.splitRuleToken["firstEvent"])
        # send HPO request
        if self.taskSpec.is_hpo_workflow():
            try:
                data = copy.copy(taskParamMap["hpoRequestData"])
                data["workload_id"] = self.taskSpec.jediTaskID
                data["is_pseudo_input"] = True
                req = {
                    "requester": "panda",
                    "request_type": idds.common.constants.RequestType.HyperParameterOpt,
                    "transform_tag": idds.common.constants.RequestType.HyperParameterOpt.value,
                    "status": idds.common.constants.RequestStatus.New,
                    "priority": 0,
                    "lifetime": 30,
                    "request_metadata": data,
                }
                c = iDDS_Client(idds.common.utils.get_rest_host())
                self.tmpLog.debug(f"req {str(req)}")
                ret = c.add_request(**req)
                self.tmpLog.debug(f"got requestID={str(ret)}")
            except Exception as e:
                errStr = f"iDDS failed with {str(e)}"
                raise JediException.ExternalTempError(errStr)

        return

    # replace placeholder with dict provided by prepro job
    def replacePlaceHolders(self, paramItem, placeHolderName, newValue):
        if isinstance(paramItem, dict):
            # loop over all dict params
            for tmpParName, tmpParVal in paramItem.items():
                if tmpParVal == placeHolderName:
                    # replace placeholder
                    paramItem[tmpParName] = newValue
                elif isinstance(tmpParVal, dict) or isinstance(tmpParVal, list):
                    # recursive execution
                    self.replacePlaceHolders(tmpParVal, placeHolderName, newValue)
        elif isinstance(paramItem, list):
            # loop over all list items
            for tmpItem in paramItem:
                self.replacePlaceHolders(tmpItem, placeHolderName, newValue)

    # refinement procedure for preprocessing
    def doPreProRefine(self, taskParamMap):
        # no preprocessing
        if "preproSpec" not in taskParamMap:
            return None, taskParamMap
        # already preprocessed
        if self.taskSpec.checkPreProcessed():
            # get replaced task params
            tmpStat, tmpJsonStr = self.taskBufferIF.getPreprocessMetadata_JEDI(self.taskSpec.jediTaskID)
            try:
                # replace placeholders
                replaceParams = RefinerUtils.decodeJSON(tmpJsonStr)
                self.tmpLog.debug("replace placeholders with " + str(replaceParams))
                for tmpKey, tmpVal in replaceParams.items():
                    self.replacePlaceHolders(taskParamMap, tmpKey, tmpVal)
            except Exception:
                errtype, errvalue = sys.exc_info()[:2]
                self.tmpLog.error(f"{self.__class__.__name__} failed to get additional task params with {errtype.__name__}:{errvalue}")
                return False, taskParamMap
            # succeeded
            self.updatedTaskParams = taskParamMap
            return None, taskParamMap
        # make dummy dataset to keep track of preprocessing
        datasetSpec = JediDatasetSpec()
        datasetSpec.datasetName = f"panda.pp.in.{uuid.uuid4()}.{self.taskSpec.jediTaskID}"
        datasetSpec.jediTaskID = self.taskSpec.jediTaskID
        datasetSpec.type = "pp_input"
        datasetSpec.vo = self.taskSpec.vo
        datasetSpec.nFiles = 1
        datasetSpec.nFilesUsed = 0
        datasetSpec.nFilesToBeUsed = 1
        datasetSpec.nFilesFinished = 0
        datasetSpec.nFilesFailed = 0
        datasetSpec.nFilesOnHold = 0
        datasetSpec.status = "ready"
        self.inMasterDatasetSpec.append(datasetSpec)
        # make file
        fileSpec = JediFileSpec()
        fileSpec.jediTaskID = datasetSpec.jediTaskID
        fileSpec.type = datasetSpec.type
        fileSpec.status = "ready"
        fileSpec.lfn = "pseudo_lfn"
        fileSpec.attemptNr = 0
        fileSpec.maxAttempt = 3
        fileSpec.keepTrack = 1
        datasetSpec.addFile(fileSpec)
        # make log dataset
        logDatasetSpec = JediDatasetSpec()
        logDatasetSpec.datasetName = f"panda.pp.log.{uuid.uuid4()}.{self.taskSpec.jediTaskID}"
        logDatasetSpec.jediTaskID = self.taskSpec.jediTaskID
        logDatasetSpec.type = "tmpl_pp_log"
        logDatasetSpec.streamName = "PP_LOG"
        logDatasetSpec.vo = self.taskSpec.vo
        logDatasetSpec.nFiles = 0
        logDatasetSpec.nFilesUsed = 0
        logDatasetSpec.nFilesToBeUsed = 0
        logDatasetSpec.nFilesFinished = 0
        logDatasetSpec.nFilesFailed = 0
        logDatasetSpec.nFilesOnHold = 0
        logDatasetSpec.status = "defined"
        self.outDatasetSpecList.append(logDatasetSpec)
        # make output template for log
        outTemplateMap = {
            "jediTaskID": self.taskSpec.jediTaskID,
            "serialNr": 1,
            "streamName": logDatasetSpec.streamName,
            "filenameTemplate": f"{logDatasetSpec.datasetName}._${{SN}}.log.tgz",
            "outtype": re.sub("^tmpl_", "", logDatasetSpec.type),
        }
        self.outputTemplateMap[logDatasetSpec.outputMapKey()] = [outTemplateMap]
        # set split rule to use preprocessing
        self.taskSpec.setPrePro()
        # set task status
        self.taskSpec.status = "topreprocess"
        # return
        return True, taskParamMap

    # set split rule
    def setSplitRule(self, taskParamMap, keyName, valName):
        if taskParamMap is not None:
            if keyName not in taskParamMap:
                return
            tmpStr = f"{valName}={taskParamMap[keyName]}"
        else:
            tmpStr = f"{valName}={keyName}"
        if self.taskSpec.splitRule in [None, ""]:
            self.taskSpec.splitRule = tmpStr
        else:
            tmpMatch = re.search(valName + "=(-*\d+)(,-*\d+)*", self.taskSpec.splitRule)
            if tmpMatch is None:
                # append
                self.taskSpec.splitRule += f",{tmpStr}"
            else:
                # replace
                self.taskSpec.splitRule = self.taskSpec.splitRule.replace(tmpMatch.group(0), tmpStr)
        return

    # get parameters for event service merging
    def getParamsForEventServiceMerging(self, taskParamMap):
        # no event service
        if not self.taskSpec.useEventService() or self.taskSpec.on_site_merging():
            return None
        # extract parameters
        transPath = "UnDefined"
        jobParameters = "UnDefined"
        if "esmergeSpec" in taskParamMap:
            if "transPath" in taskParamMap["esmergeSpec"]:
                transPath = taskParamMap["esmergeSpec"]["transPath"]
            if "jobParameters" in taskParamMap["esmergeSpec"]:
                jobParameters = jobParameters["esmergeSpec"]["jobParameters"]
        # return
        return "<PANDA_ESMERGE_TRF>" + transPath + "</PANDA_ESMERGE_TRF>" + "<PANDA_ESMERGE_JOBP>" + jobParameters + "</PANDA_ESMERGE_JOBP>"


Interaction.installSC(TaskRefinerBase)



import copy
import math
import os
import re

from pandaserver.taskbuffer import JobUtils


# get effective file size
def getEffectiveFileSize(fsize, startEvent, endEvent, nEvents):
    inMB = 1024 * 1024
    if fsize in [None, 0]:
        # use dummy size for pseudo input
        effectiveFsize = inMB
    elif nEvents is not None and startEvent is not None and endEvent is not None:
        # take event range into account
        effectiveFsize = int(float(fsize) * float(endEvent - startEvent + 1) / float(nEvents))
    else:
        effectiveFsize = fsize
    # use dummy size if input is too small
    if effectiveFsize == 0:
        effectiveFsize = inMB
    # in MB
    effectiveFsize = float(effectiveFsize) / inMB
    # return
    return effectiveFsize


# get effective number of events
def getEffectiveNumEvents(startEvent, endEvent, nEvents):
    if endEvent is not None and startEvent is not None:
        evtCounts = endEvent - startEvent + 1
        if evtCounts > 0:
            return evtCounts
        return 1
    if nEvents is not None:
        return nEvents
    return 1


# get memory usage
def getMemoryUsage():
    try:
        t = open(f"/proc/{os.getpid()}/status")
        v = t.read()
        t.close()
        value = 0
        for line in v.split("\n"):
            if line.startswith("VmRSS"):
                items = line.split()
                value = int(items[1])
                if items[2] in ["kB", "KB"]:
                    value /= 1024
                elif items[2] in ["mB", "MB"]:
                    pass
                break
        return value
    except Exception:
        return None


# check process
def checkProcess(pid):
    return os.path.exists(f"/proc/{pid}/status")


# offset for walltime
wallTimeOffset = 10 * 60


# add offset to walltime
def addOffsetToWalltime(oldWalltime):
    if oldWalltime > 0:
        # add offset of 10min
        oldWalltime += wallTimeOffset
    return oldWalltime


# reduce offset from walltime
def reduceOffsetFromWalltime(oldWalltime):
    if oldWalltime > 0:
        # add offset of 10min
        oldWalltime -= wallTimeOffset
        if oldWalltime < 0:
            oldWalltime = 0
    return oldWalltime


# convert config parameters
def convert_config_params(itemStr):
    items = itemStr.split(":")
    newItems = []
    for item in items:
        if item == "":
            newItems.append(None)
        elif "," in item:
            newItems.append(item.split(","))
        else:
            try:
                newItems.append(int(item))
            except Exception:
                newItems.append(item)
    return newItems


# parse init params
def parse_init_params(par):
    if isinstance(par, list):
        return par
    try:
        return par.split("|")
    except Exception:
        return [par]


# get config param for vo and prodSourceLabel
def getConfigParam(configStr, vo, sourceLabel):
    try:
        for tmpConfigStr in configStr.split(","):
            items = configStr.split(":")
            vos = items[0].split("|")
            sourceLabels = items[1].split("|")
            if vo not in ["", "any"] and vo not in vos and None not in vos and "any" not in vos and "" not in vos:
                continue
            if (
                sourceLabel not in ["", "any"]
                and sourceLabel not in sourceLabels
                and None not in sourceLabels
                and "any" not in sourceLabels
                and "" not in sourceLabels
            ):
                continue
            return ",".join(items[2:])
    except Exception:
        pass
    return None


# get percentile until numpy 1.5.X becomes available
def percentile(inList, percent, idMap):
    inList = sorted(copy.copy(inList))
    k = (len(inList) - 1) * float(percent) / 100
    f = math.floor(k)
    c = math.ceil(k)
    if f == c:
        retVal = inList[int(f)]
        return retVal, [retVal]
    val0 = inList[int(f)]
    val1 = inList[int(c)]
    d0 = val0 * (c - k)
    d1 = val1 * (k - f)
    retVal = d0 + d1
    return retVal, [val0, val1]


# get min ram count for job
def getJobMinRamCount(taskSpec, inputChunk, siteSpec, coreCount):
    minRamCount = inputChunk.getMaxRamCount()
    if inputChunk.isMerging:
        minRamUnit = "MB"
    else:
        minRamUnit = taskSpec.ramUnit
        if minRamUnit in [None, "", "NULL"]:
            minRamUnit = "MB"
        if taskSpec.ramPerCore():
            minRamCount *= coreCount
            minRamCount += taskSpec.baseRamCount
            minRamUnit = re.sub("PerCore.*$", "", minRamUnit)
    # round up with chunks
    minRamCount = JobUtils.compensate_ram_count(minRamCount)
    return minRamCount, minRamUnit


# get max walltime and cpu count
def getJobMaxWalltime(taskSpec, inputChunk, totalMasterEvents, jobSpec, siteSpec):
    try:
        if not taskSpec.getCpuTime():
            jobSpec.maxWalltime = siteSpec.maxtime
            jobSpec.maxCpuCount = siteSpec.maxtime
        else:
            jobSpec.maxWalltime = taskSpec.getCpuTime()
            if jobSpec.maxWalltime is not None and jobSpec.maxWalltime > 0:
                jobSpec.maxWalltime *= totalMasterEvents
                if siteSpec.coreCount > 0:
                    jobSpec.maxWalltime /= float(siteSpec.coreCount)
                if siteSpec.corepower not in [0, None]:
                    jobSpec.maxWalltime /= siteSpec.corepower
            if taskSpec.cpuEfficiency not in [None, 0]:
                jobSpec.maxWalltime /= float(taskSpec.cpuEfficiency) / 100.0
            if taskSpec.baseWalltime is not None:
                jobSpec.maxWalltime += taskSpec.baseWalltime
            jobSpec.maxWalltime = int(jobSpec.maxWalltime)
            if taskSpec.useHS06():
                jobSpec.maxCpuCount = jobSpec.maxWalltime
    except Exception:
        pass


# use direct IO for job
def use_direct_io_for_job(task_spec, site_spec, input_chunk):
    # not for merging
    if input_chunk and input_chunk.isMerging:
        return False
    # always
    if site_spec.always_use_direct_io():
        return True
    # force copy-to-scratch
    if task_spec.useLocalIO():
        return False
    # depends on task and site specs
    if task_spec.allowInputLAN() is not None and site_spec.isDirectIO():
        return True
    return False



import json

from pandacommon.kafkapublisher.KafkaPublisher import KafkaPublisher
from pandacommon.pandalogger import logger_utils
from pandajedi.jedimsgprocessor.base_msg_processor import BaseMsgProcPlugin

# Logger
base_logger = logger_utils.setup_logger(__name__.split(".")[-1])


# Kafka message processing plugin
class KafkaMsgProcPlugin(BaseMsgProcPlugin):
    def initialize(self):
        """
        initialize plugin instance, run once before loop in thread
        """
        self.publisher = KafkaPublisher()

    def process(self, msg_obj, decoded_data=None):
        # logger
        tmp_log = logger_utils.make_logger(base_logger, token=self.get_pid(), method_name="process")

        # start
        tmp_log.info("start")
        tmp_log.debug(f"sub_id={msg_obj.sub_id} ; msg_id={msg_obj.msg_id}")

        # Parse and access the message content from msg_obj.data
        if decoded_data is None:
            # json decode
            try:
                message_content = json.loads(msg_obj.data)
            except Exception as e:
                err_str = f"failed to parse message json {msg_obj.data} , skipped. {e.__class__.__name__} : {e}"
                tmp_log.error(err_str)
                raise
        else:
            message_content = decoded_data

        # Publish the message to Kafka
        self.publisher.publish_message(message_content)
        tmp_log.debug(f"sent {message_content}")
        tmp_log.info("done")

    def terminate(self):
        self.publisher.close()



from pandacommon.pandalogger import logger_utils
from pandajedi.jediconfig import jedi_config
from pandajedi.jedidaemons.utils import DaemonMaster

# logger
base_logger = logger_utils.setup_logger(__name__.split(".")[-1])


# launch
def launcher(taskBufferIF, ddmIF):
    tmp_log = logger_utils.make_logger(base_logger, method_name="launcher")
    tmp_log.debug("start")
    try:
        jedi_config.daemon.config
    except Exception as e:
        tmp_log.error(f"failed to read config json file; should not happen... {e.__class__.__name__}: {e}")
        raise e
    # whether to run daemons
    if not getattr(jedi_config.daemon, "enable", False):
        tmp_log.debug("daemon disabled ; skipped")
        return
    # parameters
    n_workers = getattr(jedi_config.daemon, "n_proc", 1)
    worker_lifetime = getattr(jedi_config.daemon, "proc_lifetime", 28800)
    # start
    agent = DaemonMaster(logger=tmp_log, n_workers=n_workers, worker_lifetime=worker_lifetime, tbuf=taskBufferIF, ddmif=ddmIF)
    agent.run()



import traceback

# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandajedi.jedicore import Interaction
from pandajedi.jedicore.MsgWrapper import MsgWrapper

from .TaskSetupperBase import TaskSetupperBase

logger = PandaLogger().getLogger(__name__.split(".")[-1])


# task setup for general purpose
class SimpleTaskSetupper(TaskSetupperBase):
    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        TaskSetupperBase.__init__(self, taskBufferIF, ddmIF)

    # main to setup task
    def doSetup(self, taskSpec, datasetToRegister, pandaJobs):
        # make logger
        tmpLog = MsgWrapper(logger, f"< jediTaskID={taskSpec.jediTaskID} >")
        tmpLog.info(f"start label={taskSpec.prodSourceLabel} taskType={taskSpec.taskType}")
        # returns
        retFatal = self.SC_FATAL
        retOK = self.SC_SUCCEEDED
        try:
            # get DDM I/F
            ddmIF = self.ddmIF.getInterface(taskSpec.vo, taskSpec.cloud)
            # skip if DDM I/F is inactive
            if not ddmIF:
                tmpLog.info("skip due to inactive DDM I/F")
                return retOK
            # collect datasetID to register datasets/containers just in case
            for tmpPandaJob in pandaJobs:
                if not tmpPandaJob.produceUnMerge():
                    for tmpFileSpec in tmpPandaJob.Files:
                        if tmpFileSpec.type in ["output", "log"]:
                            if tmpFileSpec.datasetID not in datasetToRegister:
                                datasetToRegister.append(tmpFileSpec.datasetID)
            # register datasets
            if datasetToRegister:
                tmpLog.info(f"datasetToRegister={str(datasetToRegister)}")
                # get site mapper
                siteMapper = self.taskBufferIF.getSiteMapper()

                # loop over all datasets
                avDatasetList = []
                cnDatasetMap = {}
                ddmBackEnd = "rucio"
                for datasetID in datasetToRegister:
                    # get output and log datasets
                    tmpLog.info(f"getting datasetSpec with datasetID={datasetID}")
                    tmpStat, datasetSpec = self.taskBufferIF.getDatasetWithID_JEDI(taskSpec.jediTaskID, datasetID)
                    if not tmpStat:
                        tmpLog.error("failed to get output and log datasets")
                        return retFatal
                    if datasetSpec.isPseudo():
                        tmpLog.info("skip pseudo dataset")
                        continue

                    tmpLog.info(f"checking {datasetSpec.datasetName}")
                    # check if dataset and container are available in DDM
                    for targetName in [datasetSpec.datasetName, datasetSpec.containerName]:
                        if not targetName:
                            continue
                        if targetName in avDatasetList:
                            tmpLog.info(f"{targetName} already registered")
                            continue
                        # set lifetime
                        lifetime = None
                        # check dataset/container in DDM
                        tmpList = ddmIF.listDatasets(targetName)
                        if not tmpList:
                            # get location
                            location = None
                            locForRule = None
                            if targetName == datasetSpec.datasetName:
                                # dataset
                                tmpLog.info(f"dest={datasetSpec.destination}")
                                if datasetSpec.destination:
                                    if siteMapper.checkSite(datasetSpec.destination):
                                        location = siteMapper.getSite(datasetSpec.destination).ddm_output["default"]
                                    else:
                                        location = datasetSpec.destination
                            if locForRule is None:
                                locForRule = location
                            # set metadata
                            if targetName == datasetSpec.datasetName:
                                metaData = {}
                                metaData["task_id"] = taskSpec.jediTaskID
                                if taskSpec.campaign:
                                    metaData["campaign"] = taskSpec.campaign
                            else:
                                metaData = None
                            # register dataset/container
                            tmpLog.info(f"registering {targetName} with location={location} backend={ddmBackEnd} lifetime={lifetime} meta={str(metaData)}")
                            tmpStat = ddmIF.registerNewDataset(targetName, backEnd=ddmBackEnd, location=location, lifetime=lifetime, metaData=metaData)
                            if not tmpStat:
                                tmpLog.error(f"failed to register {targetName}")
                                return retFatal
                            # register location
                            if locForRule:
                                """
                                if taskSpec.workingGroup:
                                    userName = taskSpec.workingGroup
                                else:
                                    userName = taskSpec.userName
                                """
                                userName = None
                                activity = None
                                grouping = None
                                tmpLog.info(
                                    f"registering location={locForRule} lifetime={lifetime} days activity={activity} grouping={grouping} owner={userName}"
                                )
                                tmpStat = ddmIF.registerDatasetLocation(
                                    targetName, locForRule, owner=userName, lifetime=lifetime, backEnd=ddmBackEnd, activity=activity, grouping=grouping
                                )
                                if not tmpStat:
                                    tmpLog.error(f"failed to register location {locForRule} for {targetName}")
                                    return retFatal
                            avDatasetList.append(targetName)

                    # check if dataset is in the container
                    if datasetSpec.containerName and datasetSpec.containerName != datasetSpec.datasetName:
                        # get list of constituent datasets in the container
                        if datasetSpec.containerName not in cnDatasetMap:
                            cnDatasetMap[datasetSpec.containerName] = ddmIF.listDatasetsInContainer(datasetSpec.containerName)
                        # add dataset
                        if datasetSpec.datasetName not in cnDatasetMap[datasetSpec.containerName]:
                            tmpLog.info(f"adding {datasetSpec.datasetName} to {datasetSpec.containerName}")
                            tmpStat = ddmIF.addDatasetsToContainer(datasetSpec.containerName, [datasetSpec.datasetName], backEnd=ddmBackEnd)
                            if not tmpStat:
                                tmpLog.error(f"failed to add {datasetSpec.datasetName} to {datasetSpec.containerName}")
                                return retFatal
                            cnDatasetMap[datasetSpec.containerName].append(datasetSpec.datasetName)
                        else:
                            tmpLog.info(f"{datasetSpec.datasetName} already in {datasetSpec.containerName}")
                    # update dataset
                    datasetSpec.status = "registered"
                    self.taskBufferIF.updateDataset_JEDI(datasetSpec, {"jediTaskID": taskSpec.jediTaskID, "datasetID": datasetID})
            # return
            tmpLog.info("done")
            return retOK
        except Exception as e:
            errStr = f"doSetup failed with {str(e)}"
            tmpLog.error(errStr + traceback.format_exc())
            taskSpec.setErrDiag(errStr)
            return retFatal



import multiprocessing
import sys

from pandajedi.jedicore.JediTaskBufferInterface import JediTaskBufferInterface
from pandajedi.jediddm.DDMInterface import DDMInterface
from pandajedi.jediorder import JobGenerator

tbIF = JediTaskBufferInterface()
tbIF.setupInterface()


ddmIF = DDMInterface()
ddmIF.setupInterface()


parent_conn, child_conn = multiprocessing.Pipe()

try:
    testVO = sys.argv[1]
except Exception:
    testVO = "atlas"

try:
    testTaskType = sys.argv[2]
except Exception:
    testTaskType = "test"

try:
    execJob = False
    if sys.argv[3] == "y":
        execJob = True
except Exception:
    pass

try:
    testClouds = sys.argv[4].split(",")
except Exception:
    testClouds = [None]

print(f"{testVO} {testTaskType} {testClouds}")

gen = multiprocessing.Process(target=JobGenerator.launcher, args=(child_conn, tbIF, ddmIF, testVO, testTaskType, testClouds, False, execJob))
gen.start()



import sys
import uuid

from pandajedi.jedicore.JediDatasetSpec import JediDatasetSpec
from pandajedi.jedicore.JediTaskBufferInterface import JediTaskBufferInterface
from pandajedi.jedicore.JediTaskSpec import JediTaskSpec

tbIF = JediTaskBufferInterface()
tbIF.setupInterface()

task = JediTaskSpec()
task.jediTaskID = sys.argv[1]
task.taskName = f"pandatest.{uuid.uuid4()}"
task.status = "defined"
task.userName = "pandasrv1"
task.vo = "atlas"
task.prodSourceLabel = "managed"
task.taskPriority = 100
task.currentPriority = task.taskPriority
task.architecture = "i686-slc5-gcc43-opt"
task.transUses = "Atlas-17.2.7"
task.transHome = "AtlasProduction-17.2.8.10"
task.transPath = "Reco_trf.py"
task.workQueue_ID = 3

tbIF.insertTask_JEDI(task)


ds = JediDatasetSpec()
ds.jediTaskID = task.jediTaskID
if len(sys.argv) > 2:
    ds.datasetName = sys.argv[2]
else:
    ds.datasetName = "data12_8TeV.00214651.physics_Egamma.merge.AOD.f489_m1261"
ds.type = "input"
ds.vo = task.vo
ds.cloud = "US"
ds.streamName = "IN"
ds.status = "defined"
ds.nFiles = 0
ds.nFilesUsed = 0
ds.nFilesFinished = 0
ds.nFilesFailed = 0

st, datasetID = tbIF.insertDataset_JEDI(ds)

ds = JediDatasetSpec()
ds.jediTaskID = task.jediTaskID
ds.datasetName = "ddo.000001.Atlas.Ideal.DBRelease.v220701"
ds.type = "input"
ds.vo = task.vo
ds.cloud = "US"
ds.streamName = "DBR"
ds.status = "defined"
ds.nFiles = 0
ds.nFilesUsed = 0
ds.nFilesFinished = 0
ds.nFilesFailed = 0
ds.masterID = datasetID
ds.attributes = "repeat,nosplit"

tbIF.insertDataset_JEDI(ds)


ds = JediDatasetSpec()
ds.jediTaskID = task.jediTaskID
ds.datasetName = f"panda.jeditest.GEN.{uuid.uuid4()}"
ds.type = "output"
ds.vo = task.vo
ds.cloud = "US"
ds.streamName = "OUT"
ds.status = "defined"
ds.nFiles = 0
ds.nFilesUsed = 0
ds.nFilesFinished = 0
ds.nFilesFailed = 0

st, datasetID = tbIF.insertDataset_JEDI(ds)

tbIF.insertOutputTemplate_JEDI(
    [
        {
            "jediTaskID": task.jediTaskID,
            "datasetID": datasetID,
            "filenameTemplate": f"{ds.datasetName}.${{SN}}.pool.root",
            "serialNr": 1,
            "streamName": "OUT",
            "outtype": ds.type,
        }
    ]
)

ds = JediDatasetSpec()
ds.jediTaskID = task.jediTaskID
ds.datasetName = f"panda.jeditest.log.{uuid.uuid4()}"
ds.type = "log"
ds.vo = task.vo
ds.cloud = "US"
ds.streamName = "LOG"
ds.status = "defined"
ds.nFiles = 0
ds.nFilesUsed = 0
ds.nFilesFinished = 0
ds.nFilesFailed = 0

st, datasetID = tbIF.insertDataset_JEDI(ds)

tbIF.insertOutputTemplate_JEDI(
    [
        {
            "jediTaskID": task.jediTaskID,
            "datasetID": datasetID,
            "filenameTemplate": f"{ds.datasetName}.${{SN}}.log.tgz",
            "serialNr": 1,
            "streamName": "LOG",
            "outtype": ds.type,
        }
    ]
)

tbIF.insertJobParamsTemplate_JEDI(
    task.jediTaskID,
    'inputAODFile=${IN} maxEvents=1000 RunNumber=213816 autoConfiguration=everything preExec="from BTagging.BTaggingFlags import BTaggingFlags;BTaggingFlags.CalibrationTag="BTagCalibALL-07-02"" DBRelease=${DBR} AMITag=p1462 outputNTUP_EMBLLDNFile=${OUT}',
)



"""
dataset specification for JEDI

"""

import math
import re

from pandajedi.jediconfig import jedi_config


class JediDatasetSpec(object):
    def __str__(self):
        sb = []
        for key in self.__dict__:
            if key == "Files":
                sb.append(f"{key}='{len(self.__dict__[key])}'")
            else:
                sb.append(f"{key}='{self.__dict__[key]}'")
        return ", ".join(sb)

    def __repr__(self):
        return self.__str__()

    # attributes
    _attributes = (
        "jediTaskID",
        "datasetID",
        "datasetName",
        "containerName",
        "type",
        "creationTime",
        "modificationTime",
        "vo",
        "cloud",
        "site",
        "masterID",
        "provenanceID",
        "status",
        "state",
        "stateCheckTime",
        "stateCheckExpiration",
        "frozenTime",
        "nFiles",
        "nFilesToBeUsed",
        "nFilesUsed",
        "nFilesFinished",
        "nFilesFailed",
        "nFilesOnHold",
        "nEvents",
        "nEventsToBeUsed",
        "nEventsUsed",
        "lockedBy",
        "lockedTime",
        "attributes",
        "streamName",
        "storageToken",
        "destination",
        "templateID",
        "nFilesWaiting",
        "nFilesMissing",
    )
    # attributes which have 0 by default
    _zeroAttrs = ()
    # attributes to force update
    _forceUpdateAttrs = ("lockedBy", "lockedTime")
    # mapping between sequence and attr
    _seqAttrMap = {"datasetID": f"{jedi_config.db.schemaJEDI}.JEDI_DATASETS_ID_SEQ.nextval"}
    # token for attributes
    attrToken = {
        "allowNoOutput": "an",
        "consistencyCheck": "cc",
        "eventRatio": "er",
        "indexConsistent": "ic",
        "mergeOnly": "mo",
        "nFilesPerJob": "np",
        "num_records": "nr",
        "offset": "of",
        "objectStore": "os",
        "pseudo": "ps",
        "random": "rd",
        "reusable": "ru",
        "transient": "tr",
        "useDuplicated": "ud",
    }

    # constructor
    def __init__(self):
        # install attributes
        for attr in self._attributes:
            object.__setattr__(self, attr, None)
        # file list
        object.__setattr__(self, "Files", [])
        # map of changed attributes
        object.__setattr__(self, "_changedAttrs", {})
        # distributed
        object.__setattr__(self, "distributed", False)

    # override __setattr__ to collecte the changed attributes
    def __setattr__(self, name, value):
        oldVal = getattr(self, name)
        object.__setattr__(self, name, value)
        newVal = getattr(self, name)
        # collect changed attributes
        if oldVal != newVal or name in self._forceUpdateAttrs:
            self._changedAttrs[name] = value

    # add File to files list
    def addFile(self, fileSpec):
        # append
        self.Files.append(fileSpec)

    # reset changed attribute list
    def resetChangedList(self):
        object.__setattr__(self, "_changedAttrs", {})

    # force update
    def forceUpdate(self, name):
        if name in self._attributes:
            self._changedAttrs[name] = getattr(self, name)

    # return map of values
    def valuesMap(self, useSeq=False, onlyChanged=False):
        ret = {}
        for attr in self._attributes:
            # use sequence
            if useSeq and attr in self._seqAttrMap:
                continue
            # only changed attributes
            if onlyChanged:
                if attr not in self._changedAttrs:
                    continue
            val = getattr(self, attr)
            if val is None:
                if attr in self._zeroAttrs:
                    val = 0
                else:
                    val = None
            ret[f":{attr}"] = val
        return ret

    # pack tuple into FileSpec
    def pack(self, values):
        for i in range(len(self._attributes)):
            attr = self._attributes[i]
            val = values[i]
            object.__setattr__(self, attr, val)

    # return column names for INSERT
    def columnNames(cls, prefix=None):
        ret = ""
        for attr in cls._attributes:
            if prefix is not None:
                ret += f"{prefix}."
            ret += f"{attr},"
        ret = ret[:-1]
        return ret

    columnNames = classmethod(columnNames)

    # return expression of bind variables for INSERT
    def bindValuesExpression(cls, useSeq=True):
        ret = "VALUES("
        for attr in cls._attributes:
            if useSeq and attr in cls._seqAttrMap:
                ret += f"{cls._seqAttrMap[attr]},"
            else:
                ret += f":{attr},"
        ret = ret[:-1]
        ret += ")"
        return ret

    bindValuesExpression = classmethod(bindValuesExpression)

    # return an expression of bind variables for UPDATE to update only changed attributes
    def bindUpdateChangesExpression(self):
        ret = ""
        for attr in self._attributes:
            if attr in self._changedAttrs:
                ret += f"{attr}=:{attr},"
        ret = ret[:-1]
        ret += " "
        return ret

    # set dataset attribute
    def setDatasetAttribute(self, attr):
        if self.attributes is None:
            self.attributes = ""
        else:
            self.attributes += ","
        self.attributes += attr

    # set dataset attribute with label
    def setDatasetAttributeWithLabel(self, label):
        if label not in self.attrToken:
            return
        attr = self.attrToken[label]
        if self.attributes is None:
            self.attributes = ""
        else:
            self.attributes += ","
        self.attributes += attr

    # get the total size of files
    def getSize(self):
        totalSize = 0
        checkedList = []
        for tmpFileSpec in self.Files:
            if tmpFileSpec.lfn not in checkedList:
                totalSize += tmpFileSpec.fsize
                checkedList.append(tmpFileSpec.lfn)
        return totalSize

    # return list of status to update contents
    def statusToUpdateContents(cls):
        return ["defined", "toupdate"]

    statusToUpdateContents = classmethod(statusToUpdateContents)

    # return list of types for input
    def getInputTypes(cls):
        return ["input", "pseudo_input"]

    getInputTypes = classmethod(getInputTypes)

    # return list of types to generate jobs
    def getProcessTypes(cls):
        return cls.getInputTypes() + ["pp_input"] + cls.getMergeProcessTypes()

    getProcessTypes = classmethod(getProcessTypes)

    # return list of types for merging
    def getMergeProcessTypes(cls):
        return ["trn_log", "trn_output"]

    getMergeProcessTypes = classmethod(getMergeProcessTypes)

    # get type of unkown input
    def getUnknownInputType(cls):
        return "trn_unknown"

    getUnknownInputType = classmethod(getUnknownInputType)

    # check if JEDI needs to keep track of file usage
    def toKeepTrack(self):
        if self.isNoSplit() and self.isRepeated():
            return False
        elif self.isReusable():
            return False
        else:
            return True

    # check if it is not split
    def isNoSplit(self):
        if self.attributes is not None and "nosplit" in self.attributes:
            return True
        else:
            return False

    # check if it is repeatedly used
    def isRepeated(self):
        if self.attributes is not None and "repeat" in self.attributes:
            return True
        else:
            return False

    # check if it is randomly used
    def isRandom(self):
        if self.attributes is not None and "rd" in self.attributes.split(","):
            return True
        else:
            return False

    # check if it is reusable
    def isReusable(self):
        if self.attributes is not None and "ru" in self.attributes.split(","):
            return True
        else:
            return False

    # check if consistency is checked
    def checkConsistency(self):
        if self.attributes is not None and "cc" in self.attributes.split(","):
            return True
        else:
            return False

    # set consistency is checked
    def enableCheckConsistency(self):
        if self.attributes in [None, ""]:
            self.attributes = "cc"
        elif "cc" not in self.attributes.split(","):
            self.attributes += ",cc"

    # check if it is pseudo
    def isPseudo(self):
        if self.datasetName in ["pseudo_dataset", "seq_number"] or self.type in ["pp_input"]:
            return True
        if self.attributes is not None and self.attrToken["pseudo"] in self.attributes.split(","):
            return True
        return False

    # check if it is a many-time dataset which is treated as long-standing at T2s
    def isManyTime(self):
        if self.attributes is not None and "manytime" in self.attributes:
            return True
        else:
            return False

    # check if it is seq number
    def isSeqNumber(self):
        if self.datasetName in ["seq_number"]:
            return True
        else:
            return False

    # check if duplicated files are used
    def useDuplicatedFiles(self):
        if self.attributes is not None and ("usedup" in self.attributes or "ud" in self.attributes.split(",")):
            return True
        else:
            return False

    # check if it is a master dataset
    def isMaster(self):
        if self.masterID is None and self.type in self.getProcessTypes():
            return True
        else:
            return False

    # check if it is a master input dataset
    def isMasterInput(self):
        if self.masterID is None and self.type in self.getInputTypes():
            return True
        else:
            return False

    # remove nosplit attribute
    def remAttribute(self, attrName):
        if self.attributes is not None:
            self.attributes = re.sub(attrName, "", self.attributes)
            self.attributes = re.sub(",,", ",", self.attributes)
            self.attributes = re.sub("^,", "", self.attributes)
            self.attributes = re.sub(",$", "", self.attributes)
            if self.attributes == "":
                self.attributes = None

    # remove nosplit attribute
    def remNoSplit(self):
        self.remAttribute("nosplit")

    # remove repeat attribute
    def remRepeat(self):
        self.remAttribute("repeat")

    # get the ratio to master
    def getRatioToMaster(self):
        try:
            tmpMatch = re.search("ratio=(\d+(\.\d+)*)", self.attributes)
            if tmpMatch is not None:
                ratioStr = tmpMatch.group(1)
                try:
                    # integer
                    return int(ratioStr)
                except Exception:
                    pass
                try:
                    # float
                    return float(ratioStr)
                except Exception:
                    pass
        except Exception:
            pass
        return 1

    # get N multiplied by ratio
    def getNumMultByRatio(self, num):
        # no split
        if self.isNoSplit():
            return None
        # get ratio
        ratioVal = self.getRatioToMaster()
        # integer or float
        if isinstance(ratioVal, int):
            retVal = num * ratioVal
        else:
            retVal = float(num) * ratioVal
            retVal = int(math.ceil(retVal))
        return retVal

    # unique map key for output
    def outputMapKey(self):
        mapKey = f"{self.datasetName}#{self.provenanceID}"
        return mapKey

    # unique map key
    def uniqueMapKey(self):
        mapKey = f"{self.datasetName}#{self.datasetID}"
        return mapKey

    # set offset
    def setOffset(self, offset):
        self.setDatasetAttribute(f"{self.attrToken['offset']}={offset}")

    # get offset
    def getOffset(self):
        if self.attributes is not None:
            tmpMatch = re.search(self.attrToken["offset"] + "=(\d+)", self.attributes)
            if tmpMatch is not None:
                offset = int(tmpMatch.group(1))
                return offset
        return 0

    # set number of records
    def setNumRecords(self, n):
        self.setDatasetAttribute(f"{self.attrToken['num_records']}={n}")

    # get number of records
    def getNumRecords(self):
        if self.attributes is not None:
            for item in self.attributes.split(","):
                tmpMatch = re.search(self.attrToken["num_records"] + "=(\d+)", item)
                if tmpMatch is not None:
                    num_records = int(tmpMatch.group(1))
                    return num_records
        return None

    # set object store
    def setObjectStore(self, objectStore):
        self.setDatasetAttribute(f"{self.attrToken['objectStore']}={objectStore}")

    # get object store
    def getObjectStore(self):
        if self.attributes is not None:
            tmpMatch = re.search(self.attrToken["objectStore"] + "=([^,]+)", self.attributes)
            if tmpMatch is not None:
                return tmpMatch.group(1)
        return None

    # set the number of files per job
    def setNumFilesPerJob(self, num):
        self.setDatasetAttribute(f"{self.attrToken['nFilesPerJob']}={num}")

    # get the number of files per job
    def getNumFilesPerJob(self):
        if self.attributes is not None:
            tmpMatch = re.search(self.attrToken["nFilesPerJob"] + "=(\d+)", self.attributes)
            if tmpMatch is not None:
                num = int(tmpMatch.group(1))
                return num
        # use continuous numbers for seq_number
        if self.isSeqNumber():
            return 1
        return None

    # check if unmerged dataset
    def toMerge(self):
        if self.type.startswith("trn_"):
            return True
        return False

    # set transient
    def setTransient(self, val):
        if val is True:
            val = 1
        else:
            val = 0
        self.setDatasetAttribute(f"{self.attrToken['transient']}={val}")

    # get transient
    def getTransient(self):
        if self.attributes is not None:
            for item in self.attributes.split(","):
                tmpMatch = re.search(self.attrToken["transient"] + "=(\d+)", item)
                if tmpMatch is not None:
                    val = int(tmpMatch.group(1))
                    if val == 1:
                        return True
                    else:
                        return False
        return None

    # check if no output is allowed
    def isAllowedNoOutput(self):
        if self.attributes is not None and self.attrToken["allowNoOutput"] in self.attributes.split(","):
            return True
        else:
            return False

    # allow no output
    def allowNoOutput(self):
        if self.attributes in [None, ""]:
            items = []
        else:
            items = self.attributes.split(",")
        if self.attrToken["allowNoOutput"] not in items:
            items.append(self.attrToken["allowNoOutput"])
            self.attributes = ",".join(items)

    # check if index consistency is required
    def indexConsistent(self):
        if self.attributes is not None and self.attrToken["indexConsistent"] in self.attributes.split(","):
            return True
        else:
            return False

    # set distributed
    def setDistributed(self):
        self.distributed = True

    # reset distributed
    def reset_distributed(self):
        self.distributed = False

    # check if distributed
    def isDistributed(self):
        return self.distributed

    # set event ratio
    def setEventRatio(self, num):
        self.setDatasetAttribute(f"{self.attrToken['eventRatio']}={num}")

    # get event ratio
    def getEventRatio(self):
        if self.attributes is not None:
            for item in self.attributes.split(","):
                tmpMatch = re.search(self.attrToken["eventRatio"] + "=(\d+(\.\d+)*)", item)
                if tmpMatch is not None:
                    ratioStr = tmpMatch.group(1)
                    try:
                        # integer
                        return int(ratioStr)
                    except Exception:
                        pass
                    try:
                        # float
                        return float(ratioStr)
                    except Exception:
                        pass
        return None

    # set pseudo
    def setPseudo(self):
        if self.attributes in [None, ""]:
            items = []
        else:
            items = self.attributes.split(",")
        if self.attrToken["pseudo"] not in items:
            items.append(self.attrToken["pseudo"])
            self.attributes = ",".join(items)

    # merge only
    def is_merge_only(self):
        try:
            return self.attrToken["mergeOnly"] in self.attributes.split(",")
        except Exception:
            return False



import enum
import math
import re

"""
task specification for JEDI

"""


class JediTaskSpec(object):
    # attributes
    attributes = (
        "jediTaskID",
        "taskName",
        "status",
        "userName",
        "creationDate",
        "modificationTime",
        "startTime",
        "endTime",
        "frozenTime",
        "prodSourceLabel",
        "workingGroup",
        "vo",
        "coreCount",
        "taskType",
        "processingType",
        "taskPriority",
        "currentPriority",
        "architecture",
        "transUses",
        "transHome",
        "transPath",
        "lockedBy",
        "lockedTime",
        "termCondition",
        "splitRule",
        "walltime",
        "walltimeUnit",
        "outDiskCount",
        "outDiskUnit",
        "workDiskCount",
        "workDiskUnit",
        "ramCount",
        "ramUnit",
        "ioIntensity",
        "ioIntensityUnit",
        "workQueue_ID",
        "progress",
        "failureRate",
        "errorDialog",
        "reqID",
        "oldStatus",
        "cloud",
        "site",
        "countryGroup",
        "parent_tid",
        "eventService",
        "ticketID",
        "ticketSystemType",
        "stateChangeTime",
        "superStatus",
        "campaign",
        "mergeRamCount",
        "mergeRamUnit",
        "mergeWalltime",
        "mergeWalltimeUnit",
        "throttledTime",
        "numThrottled",
        "mergeCoreCount",
        "goal",
        "assessmentTime",
        "cpuTime",
        "cpuTimeUnit",
        "cpuEfficiency",
        "baseWalltime",
        "nucleus",
        "baseRamCount",
        "ttcRequested",
        "ttcPredicted",
        "ttcPredictionDate",
        "rescueTime",
        "requestType",
        "gshare",
        "resource_type",
        "useJumbo",
        "diskIO",
        "diskIOUnit",
        "memory_leak_core",
        "memory_leak_x2",
        "attemptNr",
        "container_name",
    )
    # attributes which have 0 by default
    _zeroAttrs = ()
    # attributes to force update
    _forceUpdateAttrs = ("lockedBy", "lockedTime")
    # mapping between sequence and attr
    _seqAttrMap = {}
    # limit length
    _limitLength = {"errorDialog": 255}
    # attribute length
    _attrLength = {"workingGroup": 32}
    # tokens for split rule
    splitRuleToken = {
        "allowEmptyInput": "AE",
        "addNthFieldToLFN": "AN",
        "allowPartialFinish": "AP",
        "altStageOut": "AT",
        "avoidVP": "AV",
        "maxCoreCount": "CC",
        "cloudAsVO": "CV",
        "ddmBackEnd": "DE",
        "disableAutoFinish": "DF",
        "disableReassign": "DI",
        "debugMode": "DM",
        "disableAutoRetry": "DR",
        "dynamicNumEvents": "DY",
        "nEsConsumers": "EC",
        "nEventsPerInput": "EI",
        "encJobParams": "EJ",
        "nEventsPerWorker": "ES",
        "firstContentsFeed": "FC",
        "failGoalUnreached": "FG",
        "fineGrainedProc": "FP",
        "firstEvent": "FT",
        "fullChain": "FU",
        "groupBoundaryID": "GB",
        "hpoWorkflow": "HO",
        "instantiateTmplSite": "IA",
        "inFilePosEvtNum": "IF",
        "ipStack": "IK",
        "allowInputLAN": "IL",
        "ignoreMissingInDS": "IM",
        "intermediateTask": "IN",
        "ipConnectivity": "IP",
        "inputPreStaging": "IS",
        "instantiateTmpl": "IT",
        "allowInputWAN": "IW",
        "noLoopingCheck": "LC",
        "useLocalIO": "LI",
        "limitedSites": "LS",
        "loadXML": "LX",
        "minCpuEfficiency": "MC",
        "messageDriven": "MD",
        "mergeEsOnOS": "ME",
        "nMaxFilesPerJob": "MF",
        "maxJumboPerSite": "MJ",
        "maxNumJobs": "MN",
        "mergeOutput": "MO",
        "multiStepExec": "MS",
        "maxWalltime": "MW",
        "maxEventsPerJob": "MX",
        "noExecStrCnv": "NC",
        "notDiscardEvents": "ND",
        "nEventsPerJob": "NE",
        "nFilesPerJob": "NF",
        "nGBPerJob": "NG",
        "noInputPooling": "NI",
        "nJumboJobs": "NJ",
        "nSitesPerJob": "NS",
        "nChunksToWait": "NT",
        "noWaitParent": "NW",
        "orderInputBy": "OI",
        "orderByLB": "OL",
        "onSiteMerging": "OM",
        "osMatching": "OS",
        "onlyTagsForFC": "OT",
        "pushStatusChanges": "PC",
        "pushJob": "PJ",
        "pfnList": "PL",
        "putLogToOS": "PO",
        "runUntilClosed": "RC",
        "registerDatasets": "RD",
        "registerEsFiles": "RE",
        "respectLB": "RL",
        "reuseSecOnDemand": "RO",
        "releasePerLB": "RP",
        "respectSplitRule": "RR",
        "randomSeed": "RS",
        "retryRamOffset": "RX",
        "retryRamStep": "RY",
        "resurrectConsumers": "SC",
        "switchEStoNormal": "SE",
        "stayOutputOnSite": "SO",
        "scoutSuccessRate": "SS",
        "useSecrets": "ST",
        "segmentedWork": "SW",
        "totNumJobs": "TJ",
        "tgtMaxOutputForNG": "TN",
        "t1Weight": "TW",
        "useBuild": "UB",
        "useJobCloning": "UC",
        "useRealNumEvents": "UE",
        "useFileAsSourceLFN": "UF",
        "usePrePro": "UP",
        "useScout": "US",
        "usePrefetcher": "UT",
        "useExhausted": "UX",
        "useZipToPin": "UZ",
        "writeInputToFile": "WF",
        "waitInput": "WI",
        "maxAttemptES": "XA",
        "decAttOnFailedES": "XF",
        "maxAttemptEsJob": "XJ",
        "nEventsPerMergeJob": "ZE",
        "nFilesPerMergeJob": "ZF",
        "nGBPerMergeJob": "ZG",
        "nMaxFilesPerMergeJob": "ZM",
    }
    # enum for preprocessing
    enum_toPreProcess = "1"
    enum_preProcessed = "2"
    enum_postPProcess = "3"
    # enum for limited sites
    enum_limitedSites = {"1": "inc", "2": "exc", "3": "incexc"}
    # enum for scout
    enum_noScout = "1"
    enum_useScout = "2"
    enum_postScout = "3"
    # enum for dataset registration
    enum_toRegisterDS = "1"
    enum_registeredDS = "2"
    # enum for IP connectivity
    enum_ipConnectivity = {"1": "full", "2": "http"}
    # enum for IP stack
    enum_ipStack = {"4": "IPv4", "6": "IPv6"}
    # enum for alternative stage-out
    enum_altStageOut = {"1": "on", "2": "off", "3": "force"}
    # enum for local direct access
    enum_inputLAN = {"1": "use", "2": "only"}
    # world cloud name
    worldCloudName = "WORLD"

    # enum for contents feeder
    class FirstContentsFeed(enum.Enum):
        TRUE = "1"
        FALSE = "0"

    # enum for useJumbo
    enum_useJumbo = {"waiting": "W", "running": "R", "pending": "P", "lack": "L", "disabled": "D"}

    # enum for input prestaging
    enum_inputPreStaging = {"use": "1", "notUse": "0"}

    # enum for full chain
    class FullChain(str, enum.Enum):
        Only = "1"
        Require = "2"
        Capable = "3"

    # enum for order input by
    class OrderInputBy(str, enum.Enum):
        eventsAlignment = "1"

    # constructor
    def __init__(self):
        # install attributes
        for attr in self.attributes:
            if attr in self._zeroAttrs:
                object.__setattr__(self, attr, 0)
            else:
                object.__setattr__(self, attr, None)
        # map of changed attributes
        object.__setattr__(self, "_changedAttrs", {})
        # template to generate job parameters
        object.__setattr__(self, "jobParamsTemplate", "")
        # associated datasets
        object.__setattr__(self, "datasetSpecList", [])
        # original error dialog
        object.__setattr__(self, "origErrorDialog", None)
        # original user name
        object.__setattr__(self, "origUserName", None)

    # override __setattr__ to collect the changed attributes
    def __setattr__(self, name, value):
        oldVal = getattr(self, name)
        if name in self._limitLength and value is not None:
            # keep original dialog
            if name == "errorDialog":
                object.__setattr__(self, "origErrorDialog", value)
            value = value[: self._limitLength[name]]
        object.__setattr__(self, name, value)
        newVal = getattr(self, name)
        # collect changed attributes
        if oldVal != newVal or name in self._forceUpdateAttrs:
            self._changedAttrs[name] = value

    # copy old attributes
    def copyAttributes(self, oldTaskSpec):
        for attr in self.attributes + ("jobParamsTemplate",):
            if "Time" in attr:
                continue
            if "Date" in attr:
                continue
            if attr in ["progress", "failureRate", "errorDialog", "status", "oldStatus", "lockedBy"]:
                continue
            self.__setattr__(attr, getattr(oldTaskSpec, attr))

    # reset changed attribute list
    def resetChangedList(self):
        object.__setattr__(self, "_changedAttrs", {})

    # reset changed attribute
    def resetChangedAttr(self, name):
        try:
            del self._changedAttrs[name]
        except Exception:
            pass

    # reserve old attributes
    def reserve_old_attributes(self):
        for attName in ["ramCount", "walltime", "cpuTime", "startTime", "cpuTimeUnit", "outDiskCount", "workDiskCount", "ioIntensity", "diskIO"]:
            self.resetChangedAttr(attName)

    # force update
    def forceUpdate(self, name):
        if name in self.attributes:
            self._changedAttrs[name] = getattr(self, name)

    # return map of values
    def valuesMap(self, useSeq=False, onlyChanged=False):
        ret = {}
        for attr in self.attributes:
            # use sequence
            if useSeq and attr in self._seqAttrMap:
                continue
            # only changed attributes
            if onlyChanged:
                if attr not in self._changedAttrs:
                    continue
            val = getattr(self, attr)
            if val is None:
                if attr in self._zeroAttrs:
                    val = 0
                else:
                    val = None
            # truncate too long values
            if attr in self._limitLength:
                if val is not None:
                    val = val[: self._limitLength[attr]]
            ret[f":{attr}"] = val
        return ret

    # pack tuple into TaskSpec
    def pack(self, values):
        for i in range(len(self.attributes)):
            attr = self.attributes[i]
            val = values[i]
            object.__setattr__(self, attr, val)

    # return column names for INSERT
    def columnNames(cls, prefix=None):
        ret = ""
        for attr in cls.attributes:
            if prefix is not None:
                ret += f"{prefix}."
            ret += f"{attr},"
        ret = ret[:-1]
        return ret

    columnNames = classmethod(columnNames)

    # return expression of bind variables for INSERT
    def bindValuesExpression(cls, useSeq=True):
        ret = "VALUES("
        for attr in cls.attributes:
            if useSeq and attr in cls._seqAttrMap:
                ret += f"{cls._seqAttrMap[attr]},"
            else:
                ret += f":{attr},"
        ret = ret[:-1]
        ret += ")"
        return ret

    bindValuesExpression = classmethod(bindValuesExpression)

    # return an expression of bind variables for UPDATE to update only changed attributes
    def bindUpdateChangesExpression(self):
        ret = ""
        for attr in self.attributes:
            if attr in self._changedAttrs:
                ret += f"{attr}=:{attr},"
        ret = ret[:-1]
        ret += " "
        return ret

    # check split rule
    def check_split_rule(self, key):
        if self.splitRule is not None:
            if re.search(self.splitRuleToken[key] + r"=(\d+)", self.splitRule):
                return True
        return False

    # get the max size per job if defined
    def getMaxSizePerJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["nGBPerJob"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                nGBPerJob = int(tmpMatch.group(1)) * 1024 * 1024 * 1024
                return nGBPerJob
        return None

    # remove nGBPerJob
    def removeMaxSizePerJob(self):
        self.removeSplitRule(self.splitRuleToken["nGBPerJob"])

    # get the max size per merge job if defined
    def getMaxSizePerMergeJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["nGBPerMergeJob"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                nGBPerJob = int(tmpMatch.group(1)) * 1024 * 1024 * 1024
                return nGBPerJob
        return None

    # get the maxnumber of files per job if defined
    def getMaxNumFilesPerJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["nMaxFilesPerJob"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # set MaxNumFilesPerJob
    def setMaxNumFilesPerJob(self, value):
        self.setSplitRule("nMaxFilesPerJob", value)

    # get the maxnumber of files per merge job if defined
    def getMaxNumFilesPerMergeJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["nMaxFilesPerMergeJob"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return 50

    # get the number of events per merge job if defined
    def getNumEventsPerMergeJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["nEventsPerMergeJob"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # check if using jumbo
    def usingJumboJobs(self):
        if self.useJumbo in self.enum_useJumbo.values() and self.useJumbo != self.enum_useJumbo["disabled"]:
            return True
        return False

    # get the number of jumbo jobs if defined
    def getNumJumboJobs(self):
        if self.usingJumboJobs() and self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["nJumboJobs"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # get the max number of jumbo jobs per site if defined
    def getMaxJumboPerSite(self):
        if self.usingJumboJobs() and self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["maxJumboPerSite"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return 1

    # get the number of sites per job
    def getNumSitesPerJob(self):
        if not self.useEventService():
            return 1
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["nSitesPerJob"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return 1

    # get the number of files per job if defined
    def getNumFilesPerJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["nFilesPerJob"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                n = int(tmpMatch.group(1))
                if self.dynamicNumEvents():
                    inn = self.get_num_events_per_input()
                    dyn = self.get_min_granularity()
                    if inn and dyn and inn > dyn:
                        n *= inn // dyn
                return n
        return None

    # remove nFilesPerJob
    def removeNumFilesPerJob(self):
        self.removeSplitRule(self.splitRuleToken["nFilesPerJob"])

    # get the number of files per merge job if defined
    def getNumFilesPerMergeJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["nFilesPerMergeJob"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # get the number of events per job if defined
    def getNumEventsPerJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["nEventsPerJob"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # get offset for random seed
    def getRndmSeedOffset(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["randomSeed"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return 0

    # get offset for first event
    def getFirstEventOffset(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["firstEvent"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return 0

    # grouping with boundaryID
    def useGroupWithBoundaryID(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["groupBoundaryID"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                gbID = int(tmpMatch.group(1))
                # 1 : input - can split,    output - free
                # 2 : input - can split,    output - mapped with provenanceID
                # 3 : input - cannot split, output - free
                # 4 : input - cannot split, output - mapped with provenanceID
                #
                # * rule for master
                # 1 : can split. one boundayID per sub chunk
                # 2 : cannot split. one boundayID per sub chunk
                # 3 : cannot split. multiple boundayIDs per sub chunk
                #
                # * rule for secodary
                # 1 : must have same boundayID. cannot split
                #
                retMap = {}
                if gbID in [1, 2]:
                    retMap["inSplit"] = 1
                else:
                    retMap["inSplit"] = 2
                if gbID in [1, 3]:
                    retMap["outMap"] = False
                else:
                    retMap["outMap"] = True
                retMap["secSplit"] = None
                return retMap
        return None

    # use build
    def useBuild(self):
        return self.check_split_rule("useBuild")

    # use sjob cloning
    def useJobCloning(self):
        return self.check_split_rule("useJobCloning")

    # get job cloning type
    def getJobCloningType(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["useJobCloning"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return tmpMatch.group(1)
        return ""

    # reuse secondary on demand
    def reuseSecOnDemand(self):
        return self.check_split_rule("reuseSecOnDemand")

    # not wait for completion of parent
    def noWaitParent(self):
        return self.check_split_rule("noWaitParent")

    # check splitRule if not wait for completion of parent
    def noWaitParentSL(cls, splitRule):
        if splitRule is not None:
            tmpMatch = re.search(cls.splitRuleToken["noWaitParent"] + "=(\d+)", splitRule)
            if tmpMatch is not None:
                return True
        return False

    noWaitParentSL = classmethod(noWaitParentSL)

    # use only limited sites
    def useLimitedSites(self):
        return self.check_split_rule("limitedSites")

    # set limited sites
    def setLimitedSites(self, policy):
        tag = None
        for tmpIdx, tmpPolicy in self.enum_limitedSites.items():
            if policy == tmpPolicy:
                tag = tmpIdx
                break
        # not found
        if tag is None:
            return
        # set
        if self.splitRule is None:
            # new
            self.splitRule = self.splitRuleToken["limitedSites"] + "=" + tag
        else:
            tmpMatch = re.search(self.splitRuleToken["limitedSites"] + "=(\d+)", self.splitRule)
            if tmpMatch is None:
                # append
                self.splitRule += "," + self.splitRuleToken["limitedSites"] + "=" + tag
            else:
                # replace
                self.splitRule = re.sub(self.splitRuleToken["limitedSites"] + "=(\d+)", self.splitRuleToken["limitedSites"] + "=" + tag, self.splitRule)

    # use local IO
    def useLocalIO(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["useLocalIO"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None and int(tmpMatch.group(1)):
                return True
        return False

    # use Event Service
    def useEventService(self, siteSpec=None):
        if self.eventService in [1, 2]:
            # check site if ES is disabled
            if self.switchEStoNormal() and siteSpec is not None and siteSpec.getJobSeed() in ["all"]:
                return False
            return True
        return False

    # get the number of events per worker for Event Service
    def getNumEventsPerWorker(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["nEventsPerWorker"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # get the number of event service consumers
    def getNumEventServiceConsumer(self):
        if not self.useEventService():
            return None
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["nEsConsumers"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # disable automatic retry
    def disableAutoRetry(self):
        return self.check_split_rule("disableAutoRetry")

    # disable reassign
    def disableReassign(self):
        return self.check_split_rule("disableReassign")

    # allow empty input
    def allowEmptyInput(self):
        return self.check_split_rule("allowEmptyInput")

    # use PFN list
    def useListPFN(self):
        return self.check_split_rule("pfnList")

    # set preprocessing
    def setPrePro(self):
        if self.splitRule is None:
            # new
            self.splitRule = self.splitRuleToken["usePrePro"] + "=" + self.enum_toPreProcess
        else:
            # append
            self.splitRule += "," + self.splitRuleToken["usePrePro"] + "=" + self.enum_toPreProcess

    # use preprocessing
    def usePrePro(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["usePrePro"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None and tmpMatch.group(1) == self.enum_toPreProcess:
                return True
        return False

    # set preprocessed
    def setPreProcessed(self):
        if self.splitRule is None:
            # new
            self.splitRule = self.splitRuleToken["usePrePro"] + "=" + self.enum_preProcessed
        else:
            tmpMatch = re.search(self.splitRuleToken["usePrePro"] + "=(\d+)", self.splitRule)
            if tmpMatch is None:
                # append
                self.splitRule += "," + self.splitRuleToken["usePrePro"] + "=" + self.enum_preProcessed
            else:
                # replace
                self.splitRule = re.sub(
                    self.splitRuleToken["usePrePro"] + "=(\d+)", self.splitRuleToken["usePrePro"] + "=" + self.enum_preProcessed, self.splitRule
                )
        return

    # check preprocessed
    def checkPreProcessed(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["usePrePro"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None and tmpMatch.group(1) == self.enum_preProcessed:
                return True
        return False

    # set post preprocess
    def setPostPreProcess(self):
        if self.splitRule is None:
            # new
            self.splitRule = self.splitRuleToken["usePrePro"] + "=" + self.enum_postPProcess
        else:
            tmpMatch = re.search(self.splitRuleToken["usePrePro"] + "=(\d+)", self.splitRule)
            if tmpMatch is None:
                # append
                self.splitRule += "," + self.splitRuleToken["usePrePro"] + "=" + self.enum_postPProcess
            else:
                # replace
                self.splitRule = re.sub(
                    self.splitRuleToken["usePrePro"] + "=(\d+)", self.splitRuleToken["usePrePro"] + "=" + self.enum_postPProcess, self.splitRule
                )
        return

    # instantiate template datasets
    def instantiateTmpl(self):
        return self.check_split_rule("instantiateTmpl")

    # instantiate template datasets at site
    def instantiateTmplSite(self):
        return self.check_split_rule("instantiateTmplSite")

    # merge output
    def mergeOutput(self):
        return self.check_split_rule("mergeOutput")

    # use random seed
    def useRandomSeed(self):
        return self.check_split_rule("randomSeed")

    # get the size of workDisk in bytes
    def getWorkDiskSize(self):
        safetyMargin = 300 * 1024 * 1024
        tmpSize = self.workDiskCount
        if tmpSize is None:
            return 0
        if self.workDiskUnit == "GB":
            tmpSize = tmpSize * 1024 * 1024 * 1024
        elif self.workDiskUnit == "MB":
            tmpSize = tmpSize * 1024 * 1024
        elif self.workDiskUnit == "kB":
            tmpSize = tmpSize * 1024
        tmpSize += safetyMargin
        return tmpSize

    # get the size of outDisk in bytes
    def getOutDiskSize(self):
        tmpSize = self.outDiskCount
        if tmpSize is None or tmpSize < 0:
            return 0
        if self.outDiskUnit is not None:
            if self.outDiskUnit.startswith("GB"):
                tmpSize = tmpSize * 1024 * 1024 * 1024
            elif self.outDiskUnit.startswith("MB"):
                tmpSize = tmpSize * 1024 * 1024
            elif self.outDiskUnit.startswith("kB"):
                tmpSize = tmpSize * 1024
        return tmpSize

    # output scales with the number of events
    def outputScaleWithEvents(self):
        if self.outDiskUnit is not None and "PerEvent" in self.outDiskUnit:
            return True
        return False

    # return list of status to update contents
    def statusToUpdateContents(cls):
        return ["defined"]

    statusToUpdateContents = classmethod(statusToUpdateContents)

    # set task status on hold
    def setOnHold(self):
        # change status
        if self.status in ["ready", "running", "merging", "scouting", "defined", "topreprocess", "preprocessing", "registered", "prepared", "rerefine"]:
            self.oldStatus = self.status
            self.status = "pending"

    # return list of status to reject external changes
    def statusToRejectExtChange(cls):
        return ["finished", "done", "prepared", "broken", "tobroken", "aborted", "toabort", "aborting", "failed", "passed"]

    statusToRejectExtChange = classmethod(statusToRejectExtChange)

    # return list of status for retry
    def statusToRetry(cls):
        return ["finished", "failed", "aborted", "exhausted"]

    statusToRetry = classmethod(statusToRetry)

    # return list of status for incexec
    def statusToIncexec(cls):
        return ["done"] + cls.statusToRetry()

    statusToIncexec = classmethod(statusToIncexec)

    # return list of status for reassign
    def statusToReassign(cls):
        return ["registered", "defined", "ready", "running", "scouting", "scouted", "pending", "assigning", "exhausted"]

    statusToReassign = classmethod(statusToReassign)

    # return list of status for Job Generator
    def statusForJobGenerator(cls):
        return ["ready", "running", "scouting", "topreprocess", "preprocessing"]

    statusForJobGenerator = classmethod(statusForJobGenerator)

    # return list of status to not pause
    def statusNotToPause(cls):
        return ["finished", "failed", "done", "aborted", "broken", "paused"]

    statusNotToPause = classmethod(statusNotToPause)

    # return mapping of command and status
    def commandStatusMap(cls):
        return {
            "kill": {"doing": "aborting", "done": "toabort"},
            "finish": {"doing": "finishing", "done": "passed"},
            "retry": {"doing": "toretry", "done": "ready"},
            "incexec": {"doing": "toincexec", "done": "rerefine"},
            "reassign": {"doing": "toreassign", "done": "reassigning"},
            "pause": {"doing": "paused", "done": "dummy"},
            "resume": {"doing": "dummy", "done": "dummy"},
            "avalanche": {"doing": "dummy", "done": "dummy"},
            "release": {"doing": "dummy", "done": "dummy"},
        }

    commandStatusMap = classmethod(commandStatusMap)

    # set task status to active
    def setInActive(self):
        # change status
        if self.status == "pending":
            self.status = self.oldStatus
            self.oldStatus = None
            self.errorDialog = None
        # reset error dialog
        self.setErrDiag(None)

    # set error dialog
    def setErrDiag(self, diag, append=False, prepend=False):
        # check if message can be encoded with UTF-8
        if diag:
            try:
                diag.encode()
            except UnicodeEncodeError:
                # remove non-ascii chars
                diag = re.sub(r"[^\x00-\x7F]+", "<non-ASCII char>", diag)
        # set error dialog
        if append is True and self.errorDialog is not None:
            if not prepend:
                self.errorDialog = f"{self.errorDialog} {diag}"
            else:
                self.errorDialog = f"{diag} {self.errorDialog}"
        elif append is None:
            # keep old log
            if self.errorDialog is None:
                self.errorDialog = diag
        else:
            self.errorDialog = diag

    # use loadXML
    def useLoadXML(self):
        return self.check_split_rule("loadXML")

    # make VOMS FQANs
    def makeFQANs(self):
        # no working group
        if self.workingGroup is not None:
            fqan = f"/{self.vo}/{self.workingGroup}/Role=production"
        else:
            if self.vo is not None:
                fqan = f"/{self.vo}/Role=NULL"
            else:
                return []
        # return
        return [fqan]

    # set split rule
    def setSplitRule(self, ruleName, ruleValue):
        if self.splitRule is None:
            # new
            self.splitRule = self.splitRuleToken[ruleName] + "=" + ruleValue
        else:
            tmpMatch = re.search(self.splitRuleToken[ruleName] + "=(\d+)", self.splitRule)
            if tmpMatch is None:
                # append
                self.splitRule += "," + self.splitRuleToken[ruleName] + "=" + ruleValue
            else:
                # replace
                self.splitRule = re.sub(self.splitRuleToken[ruleName] + "=(\d+)", self.splitRuleToken[ruleName] + "=" + ruleValue, self.splitRule)

    # remove split rule
    def removeSplitRule(self, ruleName):
        if self.splitRule is not None:
            items = self.splitRule.split(",")
            newItems = []
            for item in items:
                # remove rile
                tmpRuleName = item.split("=")[0]
                if ruleName == tmpRuleName:
                    continue
                newItems.append(item)
            self.splitRule = ",".join(newItems)

    # set to use scout
    def setUseScout(self, useFlag):
        if useFlag:
            self.setSplitRule("useScout", self.enum_useScout)
        else:
            self.setSplitRule("useScout", self.enum_noScout)

    # set post scout
    def setPostScout(self):
        self.setSplitRule("useScout", self.enum_postScout)

    # use scout
    def useScout(self, splitRule=None):
        if splitRule is None:
            splitRule = self.splitRule
        if splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["useScout"] + "=(\d+)", splitRule)
            if tmpMatch is not None and tmpMatch.group(1) == self.enum_useScout:
                return True
        return False

    # use exhausted
    def useExhausted(self):
        return self.check_split_rule("useExhausted")

    # use real number of events
    def useRealNumEvents(self):
        return self.check_split_rule("useRealNumEvents")

    # use input LFN as source for output LFN
    def useFileAsSourceLFN(self):
        return self.check_split_rule("useFileAsSourceLFN")

    # post scout
    def isPostScout(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["useScout"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None and tmpMatch.group(1) == self.enum_postScout:
                return True
        return False

    # wait until input shows up
    def waitInput(self):
        return self.check_split_rule("waitInput")

    # input prestaging
    def inputPreStaging(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["inputPreStaging"] + "=" + self.enum_inputPreStaging["use"], self.splitRule)
            if tmpMatch is not None:
                return True
        return False

    # disable input prestaging
    def disableInputPreStaging(self):
        self.setSplitRule("inputPreStaging", self.enum_inputPreStaging["notUse"])

    # set DDM backend
    def setDdmBackEnd(self, backEnd):
        if self.splitRule is None:
            # new
            self.splitRule = self.splitRuleToken["ddmBackEnd"] + "=" + backEnd
        else:
            tmpMatch = re.search(self.splitRuleToken["ddmBackEnd"] + "=([^,$]+)", self.splitRule)
            if tmpMatch is None:
                # append
                self.splitRule += "," + self.splitRuleToken["ddmBackEnd"] + "=" + backEnd
            else:
                # replace
                self.splitRule = re.sub(self.splitRuleToken["ddmBackEnd"] + "=([^,$]+)", self.splitRuleToken["ddmBackEnd"] + "=" + backEnd, self.splitRule)

    # get DDM backend
    def getDdmBackEnd(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["ddmBackEnd"] + "=([^,$]+)", self.splitRule)
            if tmpMatch is not None:
                return tmpMatch.group(1)
        return None

    # get field number to add middle name to LFN
    def getFieldNumToLFN(self):
        try:
            if self.splitRule is not None:
                tmpMatch = re.search(self.splitRuleToken["addNthFieldToLFN"] + "=([,\d]+)", self.splitRule)
                if tmpMatch is not None:
                    tmpList = tmpMatch.group(1).split(",")
                    try:
                        tmpList.remove("")
                    except Exception:
                        pass
                    return list(map(int, tmpList))
        except Exception:
            pass
        return None

    # get required success rate for scout jobs
    def getScoutSuccessRate(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["scoutSuccessRate"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # get T1 weight
    def getT1Weight(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["t1Weight"] + "=(-*\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return 0

    # respect Lumiblock boundaries
    def respectLumiblock(self):
        return self.check_split_rule("respectLB")

    # release files per Lumiblock
    def releasePerLumiblock(self):
        return self.check_split_rule("releasePerLB")

    # order by Lumiblock numbers
    def orderByLB(self):
        return self.check_split_rule("orderByLB")

    # respect split rule
    def respectSplitRule(self):
        return self.check_split_rule("respectSplitRule")

    # allow partial finish
    def allowPartialFinish(self):
        return self.check_split_rule("allowPartialFinish")

    # check if datasets should be registered
    def toRegisterDatasets(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["registerDatasets"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None and tmpMatch.group(1) == self.enum_toRegisterDS:
                return True
        return False

    # datasets were registered
    def registeredDatasets(self):
        self.setSplitRule("registerDatasets", self.enum_registeredDS)

    # set datasets to be registered
    def setToRegisterDatasets(self):
        self.setSplitRule("registerDatasets", self.enum_toRegisterDS)

    # get the max number of attempts for ES events
    def getMaxAttemptES(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["maxAttemptES"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # get the max number of attempts for ES jobs
    def getMaxAttemptEsJob(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["maxAttemptEsJob"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return self.getMaxAttemptES()

    # check attribute length
    def checkAttrLength(self):
        for attrName, attrLength in self._attrLength.items():
            attrVal = getattr(self, attrName)
            if attrVal is None:
                continue
            if len(attrVal) > attrLength:
                setattr(self, attrName, None)
                self.errorDialog = f"{attrName} is too long (actual: {len(attrVal)}, maximum: {attrLength})"
                return False
        return True

    # set IP connectivity and stack
    def setIpConnectivity(self, value):
        if not value:
            return
        values = value.split("#")
        if not values:
            return
        ipConnectivity = values[0]
        if ipConnectivity in self.enum_ipConnectivity.values():
            for tmpKey, tmpVal in self.enum_ipConnectivity.items():
                if ipConnectivity == tmpVal:
                    self.setSplitRule("ipConnectivity", tmpKey)
                    break
        if len(values) > 1:
            ipStack = values[1]
            if ipStack in self.enum_ipStack.values():
                for tmpKey, tmpVal in self.enum_ipStack.items():
                    if ipStack == tmpVal:
                        self.setSplitRule("ipStack", tmpKey)
                        break

    # get IP connectivity
    def getIpConnectivity(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["ipConnectivity"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return self.enum_ipConnectivity[tmpMatch.group(1)]
        return None

    # get IP connectivity
    def getIpStack(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["ipStack"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return self.enum_ipStack[tmpMatch.group(1)]
        return None

    # use HS06 for walltime estimation
    def useHS06(self):
        return self.cpuTimeUnit in ["HS06sPerEvent", "HS06sPerEventFixed", "mHS06sPerEvent", "mHS06sPerEventFixed"]

    # get CPU time in sec
    def getCpuTime(self):
        if not self.useHS06():
            return None
        try:
            if self.cpuTimeUnit.startswith("m"):
                return float(self.cpuTime) / 1000.0
        except Exception:
            pass
        return self.cpuTime

    # RAM scales with nCores
    def ramPerCore(self):
        return self.ramUnit in ["MBPerCore", "MBPerCoreFixed"]

    # run until input is closed
    def runUntilClosed(self):
        return self.check_split_rule("runUntilClosed")

    # stay output on site
    def stayOutputOnSite(self):
        return self.check_split_rule("stayOutputOnSite")

    # fail when goal unreached
    def failGoalUnreached(self):
        return self.check_split_rule("failGoalUnreached")

    # unset fail when goal unreached
    def unsetFailGoalUnreached(self):
        self.removeSplitRule(self.splitRuleToken["failGoalUnreached"])

    # switch ES to normal when jobs land at normal sites
    def switchEStoNormal(self):
        return self.check_split_rule("switchEStoNormal")

    # use world cloud
    def useWorldCloud(self):
        return self.cloud == self.worldCloudName

    # dynamic number of events
    def dynamicNumEvents(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["dynamicNumEvents"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return True
        return False

    # get min granularity for dynamic number of events
    def get_min_granularity(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["dynamicNumEvents"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # set alternative stage-out
    def setAltStageOut(self, value):
        if value in self.enum_altStageOut.values():
            for tmpKey, tmpVal in self.enum_altStageOut.items():
                if value == tmpVal:
                    self.setSplitRule("altStageOut", tmpKey)
                    break

    # get alternative stage-out
    def getAltStageOut(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["altStageOut"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return self.enum_altStageOut[tmpMatch.group(1)]
        return None

    # allow WAN for input access
    def allowInputWAN(self):
        return self.check_split_rule("allowInputWAN")

    # set mode for input LAN access
    def setAllowInputLAN(self, value):
        if value in self.enum_inputLAN.values():
            for tmpKey, tmpVal in self.enum_inputLAN.items():
                if value == tmpVal:
                    self.setSplitRule("allowInputLAN", tmpKey)
                    break

    # check if LAN is used for input access
    def allowInputLAN(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["allowInputLAN"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return self.enum_inputLAN[tmpMatch.group(1)]
        return None

    # put log files to OS
    def putLogToOS(self):
        return self.check_split_rule("putLogToOS")

    # merge ES on Object Store
    def mergeEsOnOS(self):
        return self.check_split_rule("mergeEsOnOS")

    # write input to file
    def writeInputToFile(self):
        return self.check_split_rule("writeInputToFile")

    # ignore missing input datasets
    def ignoreMissingInDS(self):
        return self.check_split_rule("ignoreMissingInDS")

    # suppress execute string conversion
    def noExecStrCnv(self):
        return self.check_split_rule("noExecStrCnv")

    # in-file positional event number
    def inFilePosEvtNum(self):
        return self.check_split_rule("inFilePosEvtNum")

    # register event service files
    def registerEsFiles(self):
        return self.check_split_rule("registerEsFiles")

    # disable auto finish
    def disableAutoFinish(self):
        return self.check_split_rule("disableAutoFinish")

    # reset refined attributes which may confuse they system
    def resetRefinedAttrs(self):
        self.resetChangedAttr("splitRule")
        self.resetChangedAttr("eventService")
        self.reserve_old_attributes()

    # resurrect consumers
    def resurrectConsumers(self):
        return self.check_split_rule("resurrectConsumers")

    # use prefetcher
    def usePrefetcher(self):
        return self.check_split_rule("usePrefetcher")

    # no input pooling
    def noInputPooling(self):
        return self.check_split_rule("noInputPooling")

    # get num of input chunks to wait
    def nChunksToWait(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["nChunksToWait"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # get max walltime
    def getMaxWalltime(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["maxWalltime"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1)) * 60 * 60
        return None

    # get target size of the largest output to reset NG
    def getTgtMaxOutputForNG(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["tgtMaxOutputForNG"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # not discard events
    def notDiscardEvents(self):
        return self.check_split_rule("notDiscardEvents")

    # get min CPU efficiency
    def getMinCpuEfficiency(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["minCpuEfficiency"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # decrement attemptNr of events only when failed
    def decAttOnFailedES(self):
        return self.check_split_rule("decAttOnFailedES")

    # use zip files to pin input files
    def useZipToPin(self):
        return self.check_split_rule("useZipToPin")

    # require OS matching
    def osMatching(self):
        return self.check_split_rule("osMatching")

    # architecture: sw_platform<@base_platform><#host_cpu_spec><&host_gpu_spec>
    # host_cpu_spec: architecture<-vendor<-instruction_set>>
    # host_gpu_spec: vendor<-model>

    # get SW platform
    def get_sw_platform(self):
        if self.architecture is not None:
            m = re.search("^([^@#&]*)", self.architecture)
            if m:
                return m.group(1)
        return self.architecture

    # get base platform
    def get_base_platform(self):
        if self.architecture is None or "@" not in self.architecture:
            return None
        m = re.search("@([^#&]*)", self.architecture)
        img = m.group(1)
        if img == "":
            img = None
        return img

    # get platforms
    def get_platforms(self):
        if self.architecture is not None:
            platform = self.get_sw_platform()
            base = self.get_base_platform()
            if base:
                platform += "@" + base
            return platform
        return self.architecture

    # get host CPU spec
    def get_host_cpu_spec(self):
        try:
            if not self.architecture:
                return None
            if "#" not in self.architecture:
                if re.search(r"^[\^@&]", self.architecture):
                    return None
                arch = self.architecture.split("-")[0]
                if arch:
                    return [{"arch": arch, "vendor": "*", "instr": "*"}]
                return None
            m = re.search(r"#([^\^@&]*)", self.architecture)
            spec_strs = m.group(1)
            if not spec_strs:
                return None
            # remove ()
            if spec_strs.startswith("("):
                spec_strs = spec_strs[1:]
            if spec_strs.endswith(")"):
                spec_strs = spec_strs[:-1]
            specs = []
            for spec_str in spec_strs.split("|"):
                spec_str += "-*" * (2 - spec_str.count("-"))
                if "-" not in spec_str:
                    spec_str += "-*"
                items = spec_str.split("-")
                spec = {"arch": items[0], "vendor": items[1], "instr": items[2]}
                specs.append(spec)
            return specs
        except Exception:
            return None

    # get host GPU spec
    def get_host_gpu_spec(self):
        try:
            if self.architecture is None or "&" not in self.architecture:
                return None
            m = re.search(r"&([^\^@#]*)", self.architecture)
            spec_str = m.group(1)
            if not spec_str:
                return None
            spec_str += "-*" * (1 - spec_str.count("-"))
            items = spec_str.split("-")
            spec = {"vendor": items[0], "model": items[1]}
            return spec
        except Exception:
            return None

    # HPO workflow
    def is_hpo_workflow(self):
        return self.check_split_rule("hpoWorkflow")

    # debug mode
    def is_debug_mode(self):
        return self.check_split_rule("debugMode")

    # multi-step execution
    def is_multi_step_exec(self):
        return self.check_split_rule("multiStepExec")

    # get max number of jobs
    def get_max_num_jobs(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["maxNumJobs"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # get total number of jobs
    def get_total_num_jobs(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["totNumJobs"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # use only tags for fat container
    def use_only_tags_fc(self):
        return self.check_split_rule("onlyTagsForFC")

    # avoid VP
    def avoid_vp(self):
        return self.check_split_rule("avoidVP")

    # set first contents feed
    def set_first_contents_feed(self, is_first):
        if is_first:
            self.setSplitRule("firstContentsFeed", self.FirstContentsFeed.TRUE.value)
        else:
            self.setSplitRule("firstContentsFeed", self.FirstContentsFeed.FALSE.value)

    # check if first contents feed
    def is_first_contents_feed(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["firstContentsFeed"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None and tmpMatch.group(1) == self.FirstContentsFeed.TRUE.value:
                return True
        return False

    # check if work is segmented
    def is_work_segmented(self):
        return self.check_split_rule("segmentedWork")

    # check if looping check is disabled
    def no_looping_check(self):
        return self.check_split_rule("noLoopingCheck")

    # encode job parameters
    def encode_job_params(self):
        return self.check_split_rule("encJobParams")

    # get original error dialog
    def get_original_error_dialog(self):
        if not self.origErrorDialog:
            return ""
        # remove log URL
        tmpStr = re.sub("<a href.+</a> : ", "", self.origErrorDialog)
        return tmpStr.split(". ")[-1]

    # check if secrets are used
    def use_secrets(self):
        return self.check_split_rule("useSecrets")

    # get max core count
    def get_max_core_count(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["maxCoreCount"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # push status changes
    def push_status_changes(self):
        return push_status_changes(self.splitRule)

    # use cloud as VO
    def cloud_as_vo(self):
        return self.check_split_rule("cloudAsVO")

    # push job
    def push_job(self):
        return self.check_split_rule("pushJob")

    # fine-grained process
    def is_fine_grained_process(self):
        return self.check_split_rule("fineGrainedProc")

    # on site merging
    def on_site_merging(self):
        return self.check_split_rule("onSiteMerging")

    # set full chain flag
    def set_full_chain(self, mode):
        var = None
        if mode == "only":
            var = self.FullChain.Only
        elif mode == "require":
            var = self.FullChain.Require
        elif mode == "capable":
            var = self.FullChain.Capable
        if var:
            self.setSplitRule("fullChain", var)

    # get full chain flag
    def get_full_chain(self):
        if self.splitRule:
            tmpMatch = re.search(self.splitRuleToken["fullChain"] + r"=(\d+)", self.splitRule)
            if tmpMatch:
                return tmpMatch.group(1)
        return None

    # check full chain with mode
    def check_full_chain_with_mode(self, mode):
        task_flag = self.get_full_chain()
        if mode == "only":
            if task_flag == self.FullChain.Only:
                return True
        elif mode == "require":
            if task_flag == self.FullChain.Require:
                return True
        elif mode == "capable":
            if task_flag == self.FullChain.Capable:
                return True
        return False

    # check full chain with nucleus
    def check_full_chain_with_nucleus(self, nucleus):
        if self.get_full_chain() and nucleus.get_bare_nucleus_mode():
            return True
        return False

    # get RAM for retry
    def get_ram_for_retry(self, current_ram):
        if not self.splitRule:
            return None
        tmpMatch = re.search(self.splitRuleToken["retryRamOffset"] + r"=(\d+)", self.splitRule)
        if not tmpMatch:
            return None
        offset = int(tmpMatch.group(1))
        tmpMatch = re.search(self.splitRuleToken["retryRamStep"] + r"=(\d+)", self.splitRule)
        if tmpMatch:
            step = int(tmpMatch.group(1))
        else:
            step = 0
        if not current_ram:
            return current_ram
        if current_ram < offset:
            return offset
        if not step:
            return current_ram
        return offset + math.ceil((current_ram - offset) / step) * step

    # get number of events per input
    def get_num_events_per_input(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["nEventsPerInput"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    def get_max_events_per_job(self):
        if self.splitRule is not None:
            tmpMatch = re.search(self.splitRuleToken["maxEventsPerJob"] + "=(\d+)", self.splitRule)
            if tmpMatch is not None:
                return int(tmpMatch.group(1))
        return None

    # set order input by
    def set_order_input_by(self, mode):
        var = None
        if mode == "eventsAlignment":
            var = self.OrderInputBy.eventsAlignment
        if var:
            self.setSplitRule("orderInputBy", var)

    # get full chain flag
    def order_input_by(self):
        if self.splitRule:
            tmpMatch = re.search(self.splitRuleToken["orderInputBy"] + r"=(\d+)", self.splitRule)
            if tmpMatch:
                if tmpMatch.group(1) == self.OrderInputBy.eventsAlignment:
                    return "eventsAlignment"
        return None

    # check if intermediate task
    def is_intermediate_task(self):
        return self.check_split_rule("intermediateTask")

    # check if message driven
    def is_msg_driven(self):
        return self.check_split_rule("messageDriven")


# utils


# check if push status changes without class instance
def push_status_changes(split_rule):
    if not split_rule:
        return False
    tmpMatch = re.search(JediTaskSpec.splitRuleToken["pushStatusChanges"] + r"=(\d+)", split_rule)
    if not tmpMatch or int(tmpMatch.group(1)) <= 0:
        return False
    return True



from pandajedi.jedicore.JediTaskBufferInterface import JediTaskBufferInterface

# Set up the task buffer interface
tb_if = JediTaskBufferInterface()
tb_if.setupInterface()

# Get the site mapper
site_mapper = tb_if.getSiteMapper()

# Define a random site of panda sites
panda_sites = [
    "ANALY_DESY-HH",
    "ANALY_DESY-HH_TEST",
    "WT2_Install",
    "BNL_Test_2_CE_1",
    "ANALY_IN2P3-CC-T2_RD",
    "UKI-SOUTHGRID-OX-HEP_SL6",
    "RU-Protvino-IHEP",
    "ANALY_NICS_Kraken",
    "ANALY_TRIUMF_HIMEM",
    "TESTGLEXEC",
    "ANALY_GRIF-LPNHE",
    "UKI-SCOTGRID-GLASGOW_MCORE",
    "BNL_PROD",
    "RAL-LCG2_SL6",
]

# Test the getSiteInputStorageEndpointMap function
# print tb_if.getSiteInputStorageEndpointMap(panda_sites, site_mapper)

# for panda_site in panda_sites:
#     print panda_site
#     tmp_site_spec = site_mapper.getSite('ANALY_DESY-HH')
#     print '------------------- ddm -------------------'
#     print 'ddm_input: {0}, ddm_output: {1}'.format(tmp_site_spec.ddm_input, tmp_site_spec.ddm_output)
#     print '------------------- setokens values -------------------'
#     print 'setokens_input: {0}, setokens_output: {1}'.format(tmp_site_spec.setokens_input.values(),
#                                                              tmp_site_spec.setokens_output.values())
#     print '------------------- setokens -------------------'
#     print 'setokens_input: {0}, setokens_output: {1}'.format(tmp_site_spec.setokens_input,
#                                                              tmp_site_spec.setokens_output)

from pandaserver.dataservice import DataServiceUtils  # noqa: E402

sites = sorted(site_mapper.getCloud("WORLD")["sites"])
for tmp_site_name in sites:
    print(f"tmp_site_name: {tmp_site_name}")

    tmp_site_spec = site_mapper.getSite(tmp_site_name)

    # print 'tmp_site_spec.ddm_input: {0}'.format(tmp_site_spec.ddm_input)
    # print 'tmp_site_spec.setokens_input: {0}'.format(tmp_site_spec.setokens_input.values())
    # print 'combination: {0}'.format([tmp_site_spec.ddm_input] + tmp_site_spec.setokens_input.values())

    for tmp_ddm_endpoint in [tmp_site_spec.ddm_input] + list(tmp_site_spec.setokens_input.values()):
        try:
            tmp_prefix = DataServiceUtils.getDQ2Prefix(tmp_ddm_endpoint)
            print(f"prefix: {tmp_prefix}")
        except TypeError:
            print("excepted!")

    print("-------------------")



import json
import re


# convert UTF-8 to ASCII in json dumps
def unicodeConvert(input):
    if isinstance(input, dict):
        retMap = {}
        for tmpKey, tmpVal in input.items():
            retMap[unicodeConvert(tmpKey)] = unicodeConvert(tmpVal)
        return retMap
    elif isinstance(input, list):
        retList = []
        for tmpItem in input:
            retList.append(unicodeConvert(tmpItem))
        return retList

    return input


# decode
def decodeJSON(inputStr):
    return json.loads(inputStr, object_hook=unicodeConvert)


# encode
def encodeJSON(inputMap):
    return json.dumps(inputMap)


# extract stream name
def extractStreamName(valStr):
    tmpMatch = re.search("\$\{([^\}]+)\}", valStr)
    if tmpMatch is None:
        return None
    # remove decorators
    streamName = tmpMatch.group(1)
    streamName = streamName.split("/")[0]
    return streamName


# extract output filename template and replace the value field
def extractReplaceOutFileTemplate(valStr, streamName):
    outFileTempl = valStr.split("=")[-1]
    outFileTempl = outFileTempl.replace("'", "")
    valStr = valStr.replace(outFileTempl, f"${{{streamName}}}")
    return outFileTempl, valStr


# extract file list
def extractFileList(taskParamMap, datasetName):
    baseDatasetName = datasetName.split(":")[-1]
    if "log" in taskParamMap:
        itemList = taskParamMap["jobParameters"] + [taskParamMap["log"]]
    else:
        itemList = taskParamMap["jobParameters"]
    fileList = []
    includePatt = []
    excludePatt = []
    for tmpItem in itemList:
        if (
            tmpItem["type"] == "template"
            and "dataset" in tmpItem
            and (
                (tmpItem["dataset"] == datasetName or tmpItem["dataset"] == baseDatasetName)
                or ("expandedList" in tmpItem and (datasetName in tmpItem["expandedList"] or baseDatasetName in tmpItem["expandedList"]))
            )
        ):
            if "files" in tmpItem:
                fileList = tmpItem["files"]
            if "include" in tmpItem:
                includePatt = tmpItem["include"].split(",")
            if "exclude" in tmpItem:
                excludePatt = tmpItem["exclude"].split(",")
    return fileList, includePatt, excludePatt


# append dataset
def appendDataset(taskParamMap, datasetSpec, fileList):
    # make item for dataset
    tmpItem = {}
    tmpItem["type"] = "template"
    tmpItem["value"] = ""
    tmpItem["dataset"] = datasetSpec.datasetName
    tmpItem["param_type"] = datasetSpec.type
    if fileList != []:
        tmpItem["files"] = fileList
    # append
    if "jobParameters" not in taskParamMap:
        taskParamMap["jobParameters"] = []
    taskParamMap["jobParameters"].append(tmpItem)
    return taskParamMap


# check if use random seed
def useRandomSeed(taskParamMap):
    for tmpItem in taskParamMap["jobParameters"]:
        if "value" in tmpItem:
            # get offset for random seed
            if tmpItem["type"] == "template" and tmpItem["param_type"] == "number":
                if "${RNDMSEED}" in tmpItem["value"]:
                    return True
    return False



# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger

from .JobThrottlerBase import JobThrottlerBase

logger = PandaLogger().getLogger(__name__.split(".")[-1])


# class to throttle ATLAS analysis jobs
class AtlasAnalJobThrottler(JobThrottlerBase):
    # constructor
    def __init__(self, taskBufferIF):
        JobThrottlerBase.__init__(self, taskBufferIF)
        self.logger_name = __name__.split(".")[-1]
        self.comp_name = "anal_job_throttler"
        self.app = "jedi"

    # check if throttled
    def toBeThrottled(self, vo, prodSourceLabel, cloudName, workQueue, resource_name):
        return self.toBeThrottledBase(vo, prodSourceLabel, cloudName, workQueue, resource_name, logger)

    # check if throttled
    def toBeThrottled_old(self, vo, prodSourceLabel, cloudName, workQueue, resource_name):
        # make logger
        tmpLog = MsgWrapper(logger)
        tmp_log.debug(f"start vo={vo} label={prodSourceLabel} cloud={cloudName} workQueue={workQueue.queue_name}")

        # check if unthrottled
        if not workQueue.throttled:
            tmpLog.debug("  done : unthrottled since throttled is False")
            return self.retUnThrottled

        tmpLog.debug("  done : SKIP")
        return self.retThrottled
